from ..errors import InvalidEnumValue
from ..ewsdatetime import EWSDate
from ..items import (
    CONFLICT_RESOLUTION_CHOICES,
    MESSAGE_DISPOSITION_CHOICES,
    SEND_MEETING_INVITATIONS_AND_CANCELLATIONS_CHOICES,
    SEND_ONLY,
    CalendarItem,
    Item,
)
from ..properties import ItemId
from ..util import MNS, create_element
from ..version import EXCHANGE_2013_SP1
from .common import to_item_id
from .update_folder import BaseUpdateService


class UpdateItem(BaseUpdateService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/updateitem-operation"""

    SERVICE_NAME = "UpdateItem"
    SET_FIELD_ELEMENT_NAME = "t:SetItemField"
    DELETE_FIELD_ELEMENT_NAME = "t:DeleteItemField"
    CHANGE_ELEMENT_NAME = "t:ItemChange"
    CHANGES_ELEMENT_NAME = "m:ItemChanges"
    element_container_name = f"{{{MNS}}}Items"

    def call(
        self,
        items,
        conflict_resolution,
        message_disposition,
        send_meeting_invitations_or_cancellations,
        suppress_read_receipts,
    ):
        if conflict_resolution not in CONFLICT_RESOLUTION_CHOICES:
            raise InvalidEnumValue("conflict_resolution", conflict_resolution, CONFLICT_RESOLUTION_CHOICES)
        if message_disposition not in MESSAGE_DISPOSITION_CHOICES:
            raise InvalidEnumValue("message_disposition", message_disposition, MESSAGE_DISPOSITION_CHOICES)
        if send_meeting_invitations_or_cancellations not in SEND_MEETING_INVITATIONS_AND_CANCELLATIONS_CHOICES:
            raise InvalidEnumValue(
                "send_meeting_invitations_or_cancellations",
                send_meeting_invitations_or_cancellations,
                SEND_MEETING_INVITATIONS_AND_CANCELLATIONS_CHOICES,
            )
        if message_disposition == SEND_ONLY:
            raise ValueError("Cannot send-only existing objects. Use SendItem service instead")
        return self._elems_to_objs(
            self._chunked_get_elements(
                self.get_payload,
                items=items,
                conflict_resolution=conflict_resolution,
                message_disposition=message_disposition,
                send_meeting_invitations_or_cancellations=send_meeting_invitations_or_cancellations,
                suppress_read_receipts=suppress_read_receipts,
            )
        )

    def _elem_to_obj(self, elem):
        return Item.id_from_xml(elem)

    def _update_elems(self, target, fieldnames):
        fieldnames_copy = list(fieldnames)

        if target.__class__ == CalendarItem:
            # For CalendarItem items where we update 'start' or 'end', we want to update internal timezone fields
            target.clean_timezone_fields(version=self.account.version)  # Possibly also sets timezone values
            for field_name in ("start", "end"):
                if field_name in fieldnames_copy:
                    tz_field_name = target.tz_field_for_field_name(field_name).name
                    if tz_field_name not in fieldnames_copy:
                        fieldnames_copy.append(tz_field_name)

        yield from super()._update_elems(target=target, fieldnames=fieldnames_copy)

    def _get_value(self, target, field):
        value = super()._get_value(target, field)

        if target.__class__ == CalendarItem:
            # For CalendarItem items where we update 'start' or 'end', we want to send values in the local timezone
            if field.name in ("start", "end"):
                if type(value) is EWSDate:
                    # EWS always expects a datetime
                    return target.date_to_datetime(field_name=field.name)
                tz_field_name = target.tz_field_for_field_name(field.name).name
                return value.astimezone(getattr(target, tz_field_name))

        return value

    def _target_elem(self, target):
        return to_item_id(target, ItemId)

    def get_payload(
        self,
        items,
        conflict_resolution,
        message_disposition,
        send_meeting_invitations_or_cancellations,
        suppress_read_receipts,
    ):
        # Takes a list of (Item, fieldnames) tuples where 'Item' is a instance of a subclass of Item and 'fieldnames'
        # are the attribute names that were updated.
        attrs = dict(
            ConflictResolution=conflict_resolution,
            MessageDisposition=message_disposition,
            SendMeetingInvitationsOrCancellations=send_meeting_invitations_or_cancellations,
        )
        if self.account.version.build >= EXCHANGE_2013_SP1:
            attrs["SuppressReadReceipts"] = suppress_read_receipts
        payload = create_element(f"m:{self.SERVICE_NAME}", attrs=attrs)
        payload.append(self._changes_elem(target_changes=items))
        return payload



import io
import logging
import mimetypes

from .errors import InvalidTypeError
from .fields import (
    Base64Field,
    BooleanField,
    DateTimeField,
    EWSElementField,
    FieldPath,
    IdField,
    IntegerField,
    ItemField,
    TextField,
    URIField,
)
from .properties import EWSElement, EWSMeta

log = logging.getLogger(__name__)


class AttachmentId(EWSElement):
    """'id' and 'changekey' are UUIDs generated by Exchange.

    MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/attachmentid
    """

    ELEMENT_NAME = "AttachmentId"

    ID_ATTR = "Id"
    ROOT_ID_ATTR = "RootItemId"
    ROOT_CHANGEKEY_ATTR = "RootItemChangeKey"

    id = IdField(field_uri=ID_ATTR, is_required=True)
    root_id = IdField(field_uri=ROOT_ID_ATTR)
    root_changekey = IdField(field_uri=ROOT_CHANGEKEY_ATTR)

    def __init__(self, *args, **kwargs):
        if not kwargs:
            # Allow to set attributes without keyword
            kwargs = dict(zip(self._slots_keys, args))
        super().__init__(**kwargs)


class Attachment(EWSElement, metaclass=EWSMeta):
    """Base class for FileAttachment and ItemAttachment."""

    attachment_id = EWSElementField(value_cls=AttachmentId)
    name = TextField(field_uri="Name")
    content_type = TextField(field_uri="ContentType")
    content_id = TextField(field_uri="ContentId")
    content_location = URIField(field_uri="ContentLocation")
    size = IntegerField(field_uri="Size", is_read_only=True)  # Attachment size in bytes
    last_modified_time = DateTimeField(field_uri="LastModifiedTime")
    is_inline = BooleanField(field_uri="IsInline")

    __slots__ = ("parent_item",)

    def __init__(self, **kwargs):
        self.parent_item = kwargs.pop("parent_item", None)
        super().__init__(**kwargs)

    def clean(self, version=None):
        from .items import Item

        if self.parent_item is not None and not isinstance(self.parent_item, Item):
            raise InvalidTypeError("parent_item", self.parent_item, Item)
        if self.content_type is None and self.name is not None:
            self.content_type = mimetypes.guess_type(self.name)[0] or "application/octet-stream"
        super().clean(version=version)

    def attach(self):
        from .services import CreateAttachment

        # Adds this attachment to an item and updates the changekey of the parent item
        if self.attachment_id:
            raise ValueError("This attachment has already been created")
        if not self.parent_item or not self.parent_item.account:
            raise ValueError(f"Parent item {self.parent_item} must have an account")
        item = CreateAttachment(account=self.parent_item.account).get(parent_item=self.parent_item, items=[self])
        attachment_id = item.attachment_id
        self.parent_item.changekey = attachment_id.root_changekey
        # EWS does not like receiving root_id and root_changekey on subsequent requests
        attachment_id.root_id = None
        attachment_id.root_changekey = None
        self.attachment_id = attachment_id

    def detach(self):
        from .services import DeleteAttachment

        # Deletes an attachment remotely and updates the changekey of the parent item
        if not self.attachment_id:
            raise ValueError("This attachment has not been created")
        if not self.parent_item or not self.parent_item.account:
            raise ValueError(f"Parent item {self.parent_item} must have an account")
        DeleteAttachment(account=self.parent_item.account).get(items=[self.attachment_id])
        self.parent_item = None
        self.attachment_id = None

    def __hash__(self):
        if self.attachment_id:
            return hash(self.attachment_id)
        # Be careful to avoid recursion on the back-reference to the parent item
        return hash(tuple(getattr(self, f) for f in self._slots_keys if f != "parent_item"))

    def __repr__(self):
        args_str = ", ".join(
            f"{f.name}={getattr(self, f.name)!r}" for f in self.FIELDS if f.name not in ("_item", "_content")
        )
        return f"{self.__class__.__name__}({args_str})"


class FileAttachment(Attachment):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/fileattachment"""

    ELEMENT_NAME = "FileAttachment"

    is_contact_photo = BooleanField(field_uri="IsContactPhoto")
    _content = Base64Field(field_uri="Content")

    __slots__ = ("_fp",)

    def __init__(self, **kwargs):
        kwargs["_content"] = kwargs.pop("content", None)
        super().__init__(**kwargs)
        self._fp = None

    @property
    def fp(self):
        # Return a file-like object for the content. This avoids creating multiple in-memory copies of the content.
        if self._fp is None:
            self._init_fp()
        return self._fp

    def _init_fp(self):
        # Create a file-like object for the attachment content. We try hard to reduce memory consumption so we never
        # store the full attachment content in-memory.
        if not self.parent_item or not self.parent_item.account:
            raise ValueError(f"{self.__class__.__name__} must have an account")
        self._fp = FileAttachmentIO(attachment=self)

    @property
    def content(self):
        """Return the attachment content. Stores a local copy of the content in case you want to upload the attachment
        again later.
        """
        if self.attachment_id is None:
            return self._content
        if self._content is not None:
            return self._content
        # We have an ID to the data but still haven't called GetAttachment to get the actual data. Do that now.
        with self.fp as fp:
            self._content = fp.read()
        return self._content

    @content.setter
    def content(self, value):
        """Replace the attachment content."""
        if not isinstance(value, bytes):
            raise InvalidTypeError("value", value, bytes)
        self._content = value

    @classmethod
    def from_xml(cls, elem, account):
        kwargs = {f.name: f.from_xml(elem=elem, account=account) for f in cls.FIELDS}
        kwargs["content"] = kwargs.pop("_content")
        cls._clear(elem)
        return cls(**kwargs)

    def to_xml(self, version):
        self._content = self.content  # Make sure content is available, to avoid ErrorRequiredPropertyMissing
        return super().to_xml(version=version)

    def __getstate__(self):
        # The fp does not need to be pickled
        state = {k: getattr(self, k) for k in self._slots_keys}
        del state["_fp"]
        return state

    def __setstate__(self, state):
        # Restore the fp
        for k in self._slots_keys:
            setattr(self, k, state.get(k))
        self._fp = None


class ItemAttachment(Attachment):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/itemattachment"""

    ELEMENT_NAME = "ItemAttachment"

    _item = ItemField(field_uri="Item")

    def __init__(self, **kwargs):
        kwargs["_item"] = kwargs.pop("item", None)
        super().__init__(**kwargs)

    @property
    def item(self):
        from .folders import BaseFolder
        from .services import GetAttachment

        if self.attachment_id is None:
            return self._item
        if self._item is not None:
            return self._item
        # We have an ID to the data but still haven't called GetAttachment to get the actual data. Do that now.
        if not self.parent_item or not self.parent_item.account:
            raise ValueError(f"{self.__class__.__name__} must have an account")
        additional_fields = {
            FieldPath(field=f) for f in BaseFolder.allowed_item_fields(version=self.parent_item.account.version)
        }
        attachment = GetAttachment(account=self.parent_item.account).get(
            items=[self.attachment_id],
            include_mime_content=True,
            body_type=None,
            filter_html_content=None,
            additional_fields=additional_fields,
        )
        self._item = attachment.item
        return self._item

    @item.setter
    def item(self, value):
        from .items import Item

        if not isinstance(value, Item):
            raise InvalidTypeError("value", value, Item)
        self._item = value

    @classmethod
    def from_xml(cls, elem, account):
        kwargs = {f.name: f.from_xml(elem=elem, account=account) for f in cls.FIELDS}
        kwargs["item"] = kwargs.pop("_item")
        cls._clear(elem)
        return cls(**kwargs)


class FileAttachmentIO(io.RawIOBase):
    """A BytesIO where the stream of data comes from the GetAttachment service."""

    def __init__(self, attachment):
        self._attachment = attachment
        self._stream = None
        self._overflow = None

    def readable(self):
        return True

    @property
    def closed(self):
        return self._stream is None

    def readinto(self, b):
        buf_size = len(b)  # We can't return more than l bytes
        try:
            chunk = self._overflow or next(self._stream)
        except StopIteration:
            return 0
        else:
            output, self._overflow = chunk[:buf_size], chunk[buf_size:]
            b[: len(output)] = output
            return len(output)

    def __enter__(self):
        from .services import GetAttachment

        self._stream = GetAttachment(account=self._attachment.parent_item.account).stream_file_content(
            attachment_id=self._attachment.attachment_id
        )
        self._overflow = None
        return io.BufferedReader(self, buffer_size=io.DEFAULT_BUFFER_SIZE)

    def __exit__(self, *args, **kwargs):
        self._stream = None
        self._overflow = None



from ..properties import DistinguishedFolderId, FolderId
from .base import BaseFolder, Folder
from .collections import FolderCollection
from .known_folders import (
    NON_DELETABLE_FOLDERS,
    AdminAuditLogs,
    AllContacts,
    AllItems,
    ApplicationData,
    ArchiveDeletedItems,
    ArchiveInbox,
    ArchiveMsgFolderRoot,
    ArchiveRecoverableItemsDeletions,
    ArchiveRecoverableItemsPurges,
    ArchiveRecoverableItemsRoot,
    ArchiveRecoverableItemsVersions,
    Audits,
    Birthdays,
    Calendar,
    CalendarLogging,
    CommonViews,
    Companies,
    Conflicts,
    Contacts,
    ConversationHistory,
    ConversationSettings,
    CrawlerData,
    DefaultFoldersChangeHistory,
    DeferredAction,
    DeletedItems,
    Directory,
    DlpPolicyEvaluation,
    Drafts,
    ExchangeSyncData,
    Favorites,
    Files,
    FreeBusyCache,
    FreebusyData,
    Friends,
    GALContacts,
    GraphAnalytics,
    IMContactList,
    Inbox,
    Journal,
    JunkEmail,
    LocalFailures,
    Location,
    MailboxAssociations,
    Messages,
    MsgFolderRoot,
    MyContacts,
    MyContactsExtended,
    NonDeletableFolderMixin,
    Notes,
    OrganizationalContacts,
    Outbox,
    ParkedMessages,
    PassThroughSearchResults,
    PdpProfileV2Secured,
    PeopleCentricConversationBuddies,
    PeopleConnect,
    QuickContacts,
    RecipientCache,
    RecoverableItemsDeletions,
    RecoverableItemsPurges,
    RecoverableItemsRoot,
    RecoverableItemsVersions,
    RecoveryPoints,
    Reminders,
    RSSFeeds,
    Schedule,
    SearchFolders,
    SentItems,
    ServerFailures,
    Sharing,
    Shortcuts,
    Signal,
    SkypeTeamsMessages,
    SmsAndChatsSync,
    SpoolerQueue,
    SwssItems,
    SyncIssues,
    System,
    Tasks,
    TemporarySaves,
    ToDoSearch,
    Views,
    VoiceMail,
    WellknownFolder,
    WorkingSet,
)
from .queryset import DEEP, FOLDER_TRAVERSAL_CHOICES, SHALLOW, SOFT_DELETED, FolderQuerySet, SingleFolderQuerySet
from .roots import ArchiveRoot, PublicFoldersRoot, Root, RootOfHierarchy

__all__ = [
    "AdminAuditLogs",
    "AllContacts",
    "AllItems",
    "ApplicationData",
    "ArchiveDeletedItems",
    "ArchiveInbox",
    "ArchiveMsgFolderRoot",
    "ArchiveRecoverableItemsDeletions",
    "ArchiveRecoverableItemsPurges",
    "ArchiveRecoverableItemsRoot",
    "ArchiveRecoverableItemsVersions",
    "ArchiveRoot",
    "Audits",
    "BaseFolder",
    "Birthdays",
    "Calendar",
    "CalendarLogging",
    "CommonViews",
    "Companies",
    "Conflicts",
    "Contacts",
    "ConversationHistory",
    "ConversationSettings",
    "CrawlerData",
    "DEEP",
    "DefaultFoldersChangeHistory",
    "DeferredAction",
    "DeletedItems",
    "Directory",
    "DistinguishedFolderId",
    "DlpPolicyEvaluation",
    "Drafts",
    "ExchangeSyncData",
    "FOLDER_TRAVERSAL_CHOICES",
    "Favorites",
    "Files",
    "Folder",
    "FolderCollection",
    "FolderId",
    "FolderQuerySet",
    "FreeBusyCache",
    "FreebusyData",
    "Friends",
    "GALContacts",
    "GraphAnalytics",
    "IMContactList",
    "Inbox",
    "Journal",
    "JunkEmail",
    "LocalFailures",
    "Location",
    "MailboxAssociations",
    "Messages",
    "MsgFolderRoot",
    "MyContacts",
    "MyContactsExtended",
    "NON_DELETABLE_FOLDERS",
    "NonDeletableFolderMixin",
    "Notes",
    "OrganizationalContacts",
    "Outbox",
    "ParkedMessages",
    "PassThroughSearchResults",
    "PdpProfileV2Secured",
    "PeopleCentricConversationBuddies",
    "PeopleConnect",
    "PublicFoldersRoot",
    "QuickContacts",
    "RSSFeeds",
    "RecipientCache",
    "RecoverableItemsDeletions",
    "RecoverableItemsPurges",
    "RecoverableItemsRoot",
    "RecoverableItemsVersions",
    "RecoveryPoints",
    "Reminders",
    "Root",
    "RootOfHierarchy",
    "SHALLOW",
    "SOFT_DELETED",
    "Schedule",
    "SearchFolders",
    "SentItems",
    "ServerFailures",
    "Sharing",
    "Shortcuts",
    "Signal",
    "SingleFolderQuerySet",
    "SkypeTeamsMessages",
    "SmsAndChatsSync",
    "SpoolerQueue",
    "SwssItems",
    "SyncIssues",
    "System",
    "Tasks",
    "TemporarySaves",
    "ToDoSearch",
    "Views",
    "VoiceMail",
    "WellknownFolder",
    "WorkingSet",
]



from ..errors import ResponseMessageError
from ..util import MNS, create_element
from .common import EWSAccountService, item_ids_element


class ExportItems(EWSAccountService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/exportitems-operation"""

    ERRORS_TO_CATCH_IN_RESPONSE = ResponseMessageError
    SERVICE_NAME = "ExportItems"
    element_container_name = f"{{{MNS}}}Data"

    def call(self, items):
        return self._elems_to_objs(self._chunked_get_elements(self.get_payload, items=items))

    def _elem_to_obj(self, elem):
        return elem.text  # All we want is the 64bit string in the 'Data' tag

    def get_payload(self, items):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        payload.append(item_ids_element(items=items, version=self.account.version))
        return payload

    # We need to override this since ExportItemsResponseMessage is formatted a
    # little bit differently. .
    @classmethod
    def _get_elements_in_container(cls, container):
        return [container]



from ..properties import FreeBusyView
from ..util import MNS, create_element, set_xml_value
from .common import EWSService


class GetUserAvailability(EWSService):
    """Get detailed availability information for a list of users.
    MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getuseravailability-operation
    """

    SERVICE_NAME = "GetUserAvailability"

    def call(self, timezone, mailbox_data, free_busy_view_options):
        # TODO: Also supports SuggestionsViewOptions, see
        #  https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/suggestionsviewoptions
        return self._elems_to_objs(
            self._get_elements(
                payload=self.get_payload(
                    timezone=timezone, mailbox_data=mailbox_data, free_busy_view_options=free_busy_view_options
                )
            )
        )

    def _elem_to_obj(self, elem):
        return FreeBusyView.from_xml(elem=elem, account=None)

    def get_payload(self, timezone, mailbox_data, free_busy_view_options):
        payload = create_element(f"m:{self.SERVICE_NAME}Request")
        set_xml_value(payload, timezone, version=self.protocol.version)
        mailbox_data_array = create_element("m:MailboxDataArray")
        set_xml_value(mailbox_data_array, mailbox_data, version=self.protocol.version)
        payload.append(mailbox_data_array)
        return set_xml_value(payload, free_busy_view_options, version=self.protocol.version)

    @staticmethod
    def _response_messages_tag():
        return f"{{{MNS}}}FreeBusyResponseArray"

    @classmethod
    def _response_message_tag(cls):
        return f"{{{MNS}}}FreeBusyResponse"

    def _get_elements_in_response(self, response):
        for msg in response:
            container_or_exc = self._get_element_container(message=msg.find(f"{{{MNS}}}ResponseMessage"))
            if isinstance(container_or_exc, Exception):
                yield container_or_exc
            else:
                yield from self._get_elements_in_container(container=msg)

    @classmethod
    def _get_elements_in_container(cls, container):
        return [container.find(f"{{{MNS}}}FreeBusyView")]



from ..errors import AutoDiscoverFailed, ErrorNonExistentMailbox
from ..fields import (
    BooleanField,
    BuildField,
    Choice,
    ChoiceField,
    EmailAddressField,
    EWSElementField,
    IntegerField,
    OnOffField,
    ProtocolListField,
    TextField,
)
from ..properties import EWSElement
from ..transport import BASIC, CBA, DEFAULT_ENCODING, GSSAPI, NOAUTH, NTLM, SSPI
from ..util import AUTODISCOVER_BASE_NS, AUTODISCOVER_REQUEST_NS
from ..util import AUTODISCOVER_RESPONSE_NS as RNS
from ..util import ParseError, add_xml_child, create_element, is_xml, to_xml, xml_to_str
from ..version import Version


class AutodiscoverBase(EWSElement):
    NAMESPACE = RNS


class User(AutodiscoverBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/user-pox"""

    ELEMENT_NAME = "User"

    display_name = TextField(field_uri="DisplayName", namespace=RNS)
    legacy_dn = TextField(field_uri="LegacyDN", namespace=RNS)
    deployment_id = TextField(field_uri="DeploymentId", namespace=RNS)  # GUID format
    autodiscover_smtp_address = EmailAddressField(field_uri="AutoDiscoverSMTPAddress", namespace=RNS)


class IntExtUrlBase(AutodiscoverBase):
    external_url = TextField(field_uri="ExternalUrl", namespace=RNS)
    internal_url = TextField(field_uri="InternalUrl", namespace=RNS)


class AddressBook(IntExtUrlBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/addressbook-pox"""

    ELEMENT_NAME = "AddressBook"


class MailStore(IntExtUrlBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/mailstore-pox"""

    ELEMENT_NAME = "MailStore"


class NetworkRequirements(AutodiscoverBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/networkrequirements-pox"""

    ELEMENT_NAME = "NetworkRequirements"

    ipv4_start = TextField(field_uri="IPv4Start", namespace=RNS)
    ipv4_end = TextField(field_uri="IPv4End", namespace=RNS)
    ipv6_start = TextField(field_uri="IPv6Start", namespace=RNS)
    ipv6_end = TextField(field_uri="IPv6End", namespace=RNS)


class SimpleProtocol(AutodiscoverBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/protocol-pox

    Used for the 'Internal' and 'External' elements that may contain a stripped-down version of the Protocol element.
    """

    ELEMENT_NAME = "Protocol"
    WEB = "WEB"
    EXCH = "EXCH"
    EXPR = "EXPR"
    EXHTTP = "EXHTTP"
    TYPES = (WEB, EXCH, EXPR, EXHTTP)

    type = ChoiceField(field_uri="Type", choices={Choice(c) for c in TYPES}, namespace=RNS)
    as_url = TextField(field_uri="ASUrl", namespace=RNS)


class IntExtBase(AutodiscoverBase):
    # TODO: 'OWAUrl' also has an AuthenticationMethod enum-style XML attribute with values:
    #  WindowsIntegrated, FBA, NTLM, Digest, Basic
    owa_url = TextField(field_uri="OWAUrl", namespace=RNS)
    protocol = EWSElementField(value_cls=SimpleProtocol)


class Internal(IntExtBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/internal-pox"""

    ELEMENT_NAME = "Internal"


class External(IntExtBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/external-pox"""

    ELEMENT_NAME = "External"


class Protocol(SimpleProtocol):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/protocol-pox"""

    # Attribute 'Type' is ignored here. Has a name conflict with the child element and does not seem useful.
    version = TextField(field_uri="Version", is_attribute=True, namespace=RNS)
    internal = EWSElementField(value_cls=Internal)
    external = EWSElementField(value_cls=External)
    ttl = IntegerField(field_uri="TTL", namespace=RNS, default=1)  # TTL for this autodiscover response, in hours
    server = TextField(field_uri="Server", namespace=RNS)
    server_dn = TextField(field_uri="ServerDN", namespace=RNS)
    server_version = BuildField(field_uri="ServerVersion", namespace=RNS)
    mdb_dn = TextField(field_uri="MdbDN", namespace=RNS)
    public_folder_server = TextField(field_uri="PublicFolderServer", namespace=RNS)
    port = IntegerField(field_uri="Port", namespace=RNS, min=1, max=65535)
    directory_port = IntegerField(field_uri="DirectoryPort", namespace=RNS, min=1, max=65535)
    referral_port = IntegerField(field_uri="ReferralPort", namespace=RNS, min=1, max=65535)
    ews_url = TextField(field_uri="EwsUrl", namespace=RNS)
    emws_url = TextField(field_uri="EmwsUrl", namespace=RNS)
    sharing_url = TextField(field_uri="SharingUrl", namespace=RNS)
    ecp_url = TextField(field_uri="EcpUrl", namespace=RNS)
    ecp_url_um = TextField(field_uri="EcpUrl-um", namespace=RNS)
    ecp_url_aggr = TextField(field_uri="EcpUrl-aggr", namespace=RNS)
    ecp_url_mt = TextField(field_uri="EcpUrl-mt", namespace=RNS)
    ecp_url_ret = TextField(field_uri="EcpUrl-ret", namespace=RNS)
    ecp_url_sms = TextField(field_uri="EcpUrl-sms", namespace=RNS)
    ecp_url_publish = TextField(field_uri="EcpUrl-publish", namespace=RNS)
    ecp_url_photo = TextField(field_uri="EcpUrl-photo", namespace=RNS)
    ecp_url_tm = TextField(field_uri="EcpUrl-tm", namespace=RNS)
    ecp_url_tm_creating = TextField(field_uri="EcpUrl-tmCreating", namespace=RNS)
    ecp_url_tm_hiding = TextField(field_uri="EcpUrl-tmHiding", namespace=RNS)
    ecp_url_tm_editing = TextField(field_uri="EcpUrl-tmEditing", namespace=RNS)
    ecp_url_extinstall = TextField(field_uri="EcpUrl-extinstall", namespace=RNS)
    oof_url = TextField(field_uri="OOFUrl", namespace=RNS)
    oab_url = TextField(field_uri="OABUrl", namespace=RNS)
    um_url = TextField(field_uri="UMUrl", namespace=RNS)
    ews_partner_url = TextField(field_uri="EwsPartnerUrl", namespace=RNS)
    login_name = TextField(field_uri="LoginName", namespace=RNS)
    domain_required = OnOffField(field_uri="DomainRequired", namespace=RNS)
    domain_name = TextField(field_uri="DomainName", namespace=RNS)
    spa = OnOffField(field_uri="SPA", namespace=RNS, default=True)
    auth_package = ChoiceField(
        field_uri="AuthPackage",
        namespace=RNS,
        choices={Choice(c) for c in ("basic", "kerb", "kerbntlm", "ntlm", "certificate", "negotiate", "nego2")},
    )
    cert_principal_name = TextField(field_uri="CertPrincipalName", namespace=RNS)
    ssl = OnOffField(field_uri="SSL", namespace=RNS, default=True)
    auth_required = OnOffField(field_uri="AuthRequired", namespace=RNS, default=True)
    use_pop_path = OnOffField(field_uri="UsePOPAuth", namespace=RNS)
    smtp_last = OnOffField(field_uri="SMTPLast", namespace=RNS, default=False)
    network_requirements = EWSElementField(value_cls=NetworkRequirements)
    address_book = EWSElementField(value_cls=AddressBook)
    mail_store = EWSElementField(value_cls=MailStore)

    @property
    def auth_type(self):
        # Translates 'auth_package' value to our own 'auth_type' enum vals
        if not self.auth_required:
            return NOAUTH
        if not self.auth_package:
            return None
        return {
            # Missing in list are DIGEST and OAUTH2
            "basic": BASIC,
            "kerb": GSSAPI,
            "kerbntlm": NTLM,  # Means client can chose between NTLM and GSSAPI
            "ntlm": NTLM,
            "certificate": CBA,
            "negotiate": SSPI,  # Unsure about this one
            "nego2": GSSAPI,
            "anonymous": NOAUTH,  # Seen in some docs even though it's not mentioned in MSDN
        }.get(self.auth_package.lower())


class Error(EWSElement):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/error-pox"""

    ELEMENT_NAME = "Error"
    NAMESPACE = AUTODISCOVER_BASE_NS

    id = TextField(field_uri="Id", namespace=AUTODISCOVER_BASE_NS, is_attribute=True)
    time = TextField(field_uri="Time", namespace=AUTODISCOVER_BASE_NS, is_attribute=True)
    code = TextField(field_uri="ErrorCode", namespace=AUTODISCOVER_BASE_NS)
    message = TextField(field_uri="Message", namespace=AUTODISCOVER_BASE_NS)
    debug_data = TextField(field_uri="DebugData", namespace=AUTODISCOVER_BASE_NS)


class Account(AutodiscoverBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/account-pox"""

    ELEMENT_NAME = "Account"
    REDIRECT_URL = "redirectUrl"
    REDIRECT_ADDR = "redirectAddr"
    SETTINGS = "settings"
    ACTIONS = (REDIRECT_URL, REDIRECT_ADDR, SETTINGS)

    type = ChoiceField(field_uri="AccountType", namespace=RNS, choices={Choice("email")})
    action = ChoiceField(field_uri="Action", namespace=RNS, choices={Choice(p) for p in ACTIONS})
    microsoft_online = BooleanField(field_uri="MicrosoftOnline", namespace=RNS)
    redirect_url = TextField(field_uri="RedirectURL", namespace=RNS)
    redirect_address = EmailAddressField(field_uri="RedirectAddr", namespace=RNS)
    image = TextField(field_uri="Image", namespace=RNS)  # Path to image used for branding
    service_home = TextField(field_uri="ServiceHome", namespace=RNS)  # URL to website of ISP
    protocols = ProtocolListField()
    # 'SmtpAddress' is inside the 'PublicFolderInformation' element
    public_folder_smtp_address = TextField(field_uri="SmtpAddress", namespace=RNS)

    @classmethod
    def from_xml(cls, elem, account):
        kwargs = {}
        public_folder_information = elem.find(f"{{{cls.NAMESPACE}}}PublicFolderInformation")
        for f in cls.FIELDS:
            if f.name == "public_folder_smtp_address":
                if public_folder_information is None:
                    continue
                kwargs[f.name] = f.from_xml(elem=public_folder_information, account=account)
                continue
            kwargs[f.name] = f.from_xml(elem=elem, account=account)
        cls._clear(elem)
        return cls(**kwargs)


class Response(AutodiscoverBase):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/response-pox"""

    ELEMENT_NAME = "Response"

    user = EWSElementField(value_cls=User)
    account = EWSElementField(value_cls=Account)

    @property
    def redirect_address(self):
        try:
            if self.account.action != Account.REDIRECT_ADDR:
                return None
            return self.account.redirect_address
        except AttributeError:
            return None

    @property
    def redirect_url(self):
        try:
            if self.account.action != Account.REDIRECT_URL:
                return None
            return self.account.redirect_url
        except AttributeError:
            return None

    @property
    def autodiscover_smtp_address(self):
        # AutoDiscoverSMTPAddress might not be present in the XML. In this case, use the original email address.
        try:
            if self.account.action != Account.SETTINGS:
                return None
            return self.user.autodiscover_smtp_address
        except AttributeError:
            return None

    @property
    def version(self):
        # Get the server version. Not all protocol entries have a server version so we cheat a bit and also look at the
        # other ones that point to the same endpoint.
        ews_url = self.protocol.ews_url
        for protocol in self.account.protocols:
            if not protocol.ews_url or not protocol.server_version:
                continue
            if protocol.ews_url.lower() == ews_url.lower():
                return Version(build=protocol.server_version)
        return None

    @property
    def protocol(self):
        """Return the protocol containing an EWS URL.

        A response may contain a number of possible protocol types. EXPR is meant for EWS. See
        https://techcommunity.microsoft.com/t5/blogs/blogarticleprintpage/blog-id/Exchange/article-id/16

        We allow fallback to EXCH if EXPR is not available, to support installations where EXPR is not available.

        Additionally, some responses may contain an EXPR with no EWS URL. In that case, return EXCH, if available.
        """
        protocols = {p.type: p for p in self.account.protocols if p.ews_url}
        if Protocol.EXPR in protocols:
            return protocols[Protocol.EXPR]
        if Protocol.EXCH in protocols:
            return protocols[Protocol.EXCH]
        raise ValueError(
            f"No EWS URL found in any of the available protocols: {[str(p) for p in self.account.protocols]}"
        )


class ErrorResponse(EWSElement):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/response-pox

    Like 'Response', but with a different namespace.
    """

    ELEMENT_NAME = "Response"
    NAMESPACE = AUTODISCOVER_BASE_NS

    error = EWSElementField(value_cls=Error)


class Autodiscover(EWSElement):
    ELEMENT_NAME = "Autodiscover"
    NAMESPACE = AUTODISCOVER_BASE_NS

    response = EWSElementField(value_cls=Response)
    error_response = EWSElementField(value_cls=ErrorResponse)

    @staticmethod
    def _clear(elem):
        # Parent implementation also clears the parent, but this element doesn't have one.
        elem.clear()

    @classmethod
    def from_bytes(cls, bytes_content):
        """Create an instance from response bytes. An Autodiscover request and response example is available at:
        https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/pox-autodiscover-response-for-exchange

        :param bytes_content:
        :return:
        """
        if not is_xml(bytes_content):
            raise ParseError(f"Response is not XML: {bytes_content}", "<not from file>", -1, 0)
        root = to_xml(bytes_content).getroot()  # May raise ParseError
        if root.tag != cls.response_tag():
            raise ParseError(f"Unknown root element in XML: {bytes_content}", "<not from file>", -1, 0)
        return cls.from_xml(elem=root, account=None)

    def raise_errors(self):
        # Find an error message in the response and raise the relevant exception
        try:
            errorcode = self.error_response.error.code
            message = self.error_response.error.message
            if message in ("The e-mail address cannot be found.", "The email address can't be found."):
                raise ErrorNonExistentMailbox("The SMTP address has no mailbox associated with it")
            raise AutoDiscoverFailed(f"Unknown error {errorcode}: {message}")
        except AttributeError:
            raise AutoDiscoverFailed(f"Unknown autodiscover error response: {self.error_response}")

    @staticmethod
    def payload(email):
        # Builds a full Autodiscover XML request
        payload = create_element("Autodiscover", attrs=dict(xmlns=AUTODISCOVER_REQUEST_NS))
        request = create_element("Request")
        add_xml_child(request, "EMailAddress", email)
        add_xml_child(request, "AcceptableResponseSchema", RNS)
        payload.append(request)
        return xml_to_str(payload, encoding=DEFAULT_ENCODING, xml_declaration=True)



import logging

from ..errors import EWSError, InvalidTypeError
from ..properties import Notification
from ..util import MNS, DocumentYielder, DummyResponse, create_element, get_xml_attr, get_xml_attrs
from .common import EWSAccountService, add_xml_child

log = logging.getLogger(__name__)
xml_log = logging.getLogger(f"{__name__}.xml")


class GetStreamingEvents(EWSAccountService):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getstreamingevents-operation
    """

    SERVICE_NAME = "GetStreamingEvents"
    element_container_name = f"{{{MNS}}}Notifications"
    prefer_affinity = True

    # Connection status values
    OK = "OK"
    CLOSED = "Closed"

    def __init__(self, *args, **kwargs):
        # These values are set each time call() is consumed
        self.connection_status = None
        super().__init__(*args, **kwargs)
        self.streaming = True

    def call(self, subscription_ids, connection_timeout):
        if not isinstance(connection_timeout, int):
            raise InvalidTypeError("connection_timeout", connection_timeout, int)
        if connection_timeout < 1:
            raise ValueError(f"'connection_timeout' {connection_timeout} must be a positive integer")
        # Add 60 seconds to the timeout, to allow us to always get the final message containing ConnectionStatus=Closed
        self.timeout = connection_timeout * 60 + 60
        return self._elems_to_objs(
            self._get_elements(
                payload=self.get_payload(
                    subscription_ids=subscription_ids,
                    connection_timeout=connection_timeout,
                )
            )
        )

    def _elem_to_obj(self, elem):
        return Notification.from_xml(elem=elem, account=None)

    @classmethod
    def _get_soap_parts(cls, response, **parse_opts):
        # Pass the response unaltered. We want to use our custom document yielder
        return None, response

    def _get_soap_messages(self, body, **parse_opts):
        # 'body' is actually the raw response passed on by '_get_soap_parts'. We want to continuously read the content,
        # looking for complete XML documents. When we have a full document, we want to parse it as if it was a normal
        # XML response.
        r = body
        for i, doc in enumerate(DocumentYielder(r.iter_content()), start=1):
            xml_log.debug("Response XML (docs counter: %(i)s): %(xml_response)s", dict(i=i, xml_response=doc))
            response = DummyResponse(content=doc)
            try:
                _, body = super()._get_soap_parts(response=response, **parse_opts)
            except Exception:
                r.close()  # Release memory
                raise
            # TODO: We're skipping ._update_api_version() here because we don't have access to the 'api_version' used.
            # TODO: We should be doing a lot of error handling for ._get_soap_messages().
            yield from super()._get_soap_messages(body=body, **parse_opts)
            if self.connection_status == self.CLOSED:
                # Don't wait for the TCP connection to timeout
                break

    def _get_element_container(self, message, name=None):
        error_ids_elem = message.find(f"{{{MNS}}}ErrorSubscriptionIds")
        error_ids = [] if error_ids_elem is None else get_xml_attrs(error_ids_elem, f"{{{MNS}}}SubscriptionId")
        self.connection_status = get_xml_attr(message, f"{{{MNS}}}ConnectionStatus")  # Either 'OK' or 'Closed'
        log.debug("Connection status is: %s", self.connection_status)
        # Upstream normally expects to find a 'name' tag but our response does not always have it. We still want to
        # call upstream, to have exceptions raised. Return an empty list if there is no 'name' tag and no errors.
        if message.find(name) is None:
            name = None
        try:
            res = super()._get_element_container(message=message, name=name)
        except EWSError as e:
            # When the request contains a combination of good and failing subscription IDs, notifications for the good
            # subscriptions seem to never be returned even though the XML spec allows it. This means there's no point in
            # trying to collect any notifications here and delivering a combination of errors and return values.
            if error_ids:
                e.value += f" (subscription IDs: {error_ids})"
            raise e
        return [] if name is None else res

    def get_payload(self, subscription_ids, connection_timeout):
        payload = create_element(f"m:{self.SERVICE_NAME}")
        subscriptions_elem = create_element("m:SubscriptionIds")
        for subscription_id in subscription_ids:
            add_xml_child(subscriptions_elem, "t:SubscriptionId", subscription_id)
        if not len(subscriptions_elem):
            raise ValueError("'subscription_ids' must not be empty")

        payload.append(subscriptions_elem)
        add_xml_child(payload, "m:ConnectionTimeout", connection_timeout)
        return payload



"""Stores errors specific to this package, and mirrors all the possible errors that EWS can return."""
from urllib.parse import urlparse


class MultipleObjectsReturned(Exception):
    pass


class DoesNotExist(Exception):
    pass


class InvalidEnumValue(ValueError):
    def __init__(self, field_name, value, choices):
        self.field_name = field_name
        self.value = value
        self.choices = choices
        super().__init__(str(self))

    def __str__(self):
        return f"{self.field_name!r} {self.value!r} must be one of {sorted(self.choices)}"


class InvalidTypeError(TypeError):
    def __init__(self, field_name, value, valid_type):
        self.field_name = field_name
        self.value = value
        self.valid_type = valid_type
        super().__init__(str(self))

    def __str__(self):
        return f"{self.field_name!r} {self.value!r} must be of type {self.valid_type}"


class EWSError(Exception):
    """Global error type within this module."""

    def __init__(self, value):
        super().__init__(value)
        self.value = value

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        return repr(self) == repr(other)


# Warnings
class EWSWarning(EWSError):
    pass


# Misc errors
class TransportError(EWSError):
    pass


class RateLimitError(TransportError):
    def __init__(self, value, url, status_code, total_wait):
        super().__init__(value)
        self.url = url
        self.status_code = status_code
        self.total_wait = total_wait

    def __str__(self):
        return (
            f"{self.value} (gave up after {self.total_wait:.3f} seconds. "
            f"URL {self.url} returned status code {self.status_code})"
        )


class SOAPError(TransportError):
    pass


class MalformedResponseError(TransportError):
    pass


class UnauthorizedError(EWSError):
    pass


class RedirectError(TransportError):
    def __init__(self, url):
        parsed_url = urlparse(url)
        self.url = url
        self.server = parsed_url.hostname.lower()
        self.has_ssl = parsed_url.scheme == "https"
        super().__init__(str(self))

    def __str__(self):
        return f"We were redirected to {self.url}"


class RelativeRedirect(TransportError):
    pass


class AutoDiscoverError(TransportError):
    pass


class AutoDiscoverFailed(AutoDiscoverError):
    pass


class AutoDiscoverCircularRedirect(AutoDiscoverError):
    pass


class NaiveDateTimeNotAllowed(ValueError):
    def __init__(self, local_dt):
        super().__init__()
        self.local_dt = local_dt  # An EWSDateTime instance


class UnknownTimeZone(EWSError):
    pass


class SessionPoolMinSizeReached(EWSError):
    pass


class SessionPoolMaxSizeReached(EWSError):
    pass


class ResponseMessageError(TransportError):
    pass


class CASError(EWSError):
    """EWS will sometimes return an error message in an 'X-CasErrorCode' custom HTTP header in an HTTP 500 error code.
    This exception is for those cases. The caller may want to do something with the original response, so store that.
    """

    def __init__(self, cas_error, response):
        self.cas_error = cas_error
        self.response = response
        super().__init__(str(self))

    def __str__(self):
        return f"CAS error: {self.cas_error}"


# Somewhat-authoritative list of possible response message error types from EWS. See full list at
# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/responsecode
#
class ErrorAccessDenied(ResponseMessageError):
    pass


class ErrorAccessModeSpecified(ResponseMessageError):
    pass


class ErrorAccountDisabled(ResponseMessageError):
    pass


class ErrorAddDelegatesFailed(ResponseMessageError):
    pass


class ErrorAddressSpaceNotFound(ResponseMessageError):
    pass


class ErrorADOperation(ResponseMessageError):
    pass


class ErrorADSessionFilter(ResponseMessageError):
    pass


class ErrorADUnavailable(ResponseMessageError):
    pass


class ErrorAffectedTaskOccurrencesRequired(ResponseMessageError):
    pass


class ErrorApplyConversationActionFailed(ResponseMessageError):
    pass


class ErrorAttachmentSizeLimitExceeded(ResponseMessageError):
    pass


class ErrorAutoDiscoverFailed(ResponseMessageError):
    pass


class ErrorAvailabilityConfigNotFound(ResponseMessageError):
    pass


class ErrorBatchProcessingStopped(ResponseMessageError):
    pass


class ErrorCalendarCannotMoveOrCopyOccurrence(ResponseMessageError):
    pass


class ErrorCalendarCannotUpdateDeletedItem(ResponseMessageError):
    pass


class ErrorCalendarCannotUseIdForOccurrenceId(ResponseMessageError):
    pass


class ErrorCalendarCannotUseIdForRecurringMasterId(ResponseMessageError):
    pass


class ErrorCalendarDurationIsTooLong(ResponseMessageError):
    pass


class ErrorCalendarEndDateIsEarlierThanStartDate(ResponseMessageError):
    pass


class ErrorCalendarFolderIsInvalidForCalendarView(ResponseMessageError):
    pass


class ErrorCalendarInvalidAttributeValue(ResponseMessageError):
    pass


class ErrorCalendarInvalidDayForTimeChangePattern(ResponseMessageError):
    pass


class ErrorCalendarInvalidDayForWeeklyRecurrence(ResponseMessageError):
    pass


class ErrorCalendarInvalidPropertyState(ResponseMessageError):
    pass


class ErrorCalendarInvalidPropertyValue(ResponseMessageError):
    pass


class ErrorCalendarInvalidRecurrence(ResponseMessageError):
    pass


class ErrorCalendarInvalidTimeZone(ResponseMessageError):
    pass


class ErrorCalendarIsCancelledForAccept(ResponseMessageError):
    pass


class ErrorCalendarIsCancelledForDecline(ResponseMessageError):
    pass


class ErrorCalendarIsCancelledForRemove(ResponseMessageError):
    pass


class ErrorCalendarIsCancelledForTentative(ResponseMessageError):
    pass


class ErrorCalendarIsDelegatedForAccept(ResponseMessageError):
    pass


class ErrorCalendarIsDelegatedForDecline(ResponseMessageError):
    pass


class ErrorCalendarIsDelegatedForRemove(ResponseMessageError):
    pass


class ErrorCalendarIsDelegatedForTentative(ResponseMessageError):
    pass


class ErrorCalendarIsNotOrganizer(ResponseMessageError):
    pass


class ErrorCalendarIsOrganizerForAccept(ResponseMessageError):
    pass


class ErrorCalendarIsOrganizerForDecline(ResponseMessageError):
    pass


class ErrorCalendarIsOrganizerForRemove(ResponseMessageError):
    pass


class ErrorCalendarIsOrganizerForTentative(ResponseMessageError):
    pass


class ErrorCalendarMeetingRequestIsOutOfDate(ResponseMessageError):
    pass


class ErrorCalendarOccurrenceIndexIsOutOfRecurrenceRange(ResponseMessageError):
    pass


class ErrorCalendarOccurrenceIsDeletedFromRecurrence(ResponseMessageError):
    pass


class ErrorCalendarOutOfRange(ResponseMessageError):
    pass


class ErrorCalendarViewRangeTooBig(ResponseMessageError):
    pass


class ErrorCallerIsInvalidADAccount(ResponseMessageError):
    pass


class ErrorCannotCreateCalendarItemInNonCalendarFolder(ResponseMessageError):
    pass


class ErrorCannotCreateContactInNonContactFolder(ResponseMessageError):
    pass


class ErrorCannotCreatePostItemInNonMailFolder(ResponseMessageError):
    pass


class ErrorCannotCreateTaskInNonTaskFolder(ResponseMessageError):
    pass


class ErrorCannotDeleteObject(ResponseMessageError):
    pass


class ErrorCannotDeleteTaskOccurrence(ResponseMessageError):
    pass


class ErrorCannotEmptyFolder(ResponseMessageError):
    pass


class ErrorCannotOpenFileAttachment(ResponseMessageError):
    pass


class ErrorCannotSetCalendarPermissionOnNonCalendarFolder(ResponseMessageError):
    pass


class ErrorCannotSetNonCalendarPermissionOnCalendarFolder(ResponseMessageError):
    pass


class ErrorCannotSetPermissionUnknownEntries(ResponseMessageError):
    pass


class ErrorCannotUseFolderIdForItemId(ResponseMessageError):
    pass


class ErrorCannotUseItemIdForFolderId(ResponseMessageError):
    pass


class ErrorChangeKeyRequired(ResponseMessageError):
    pass


class ErrorChangeKeyRequiredForWriteOperations(ResponseMessageError):
    pass


class ErrorClientDisconnected(ResponseMessageError):
    pass


class ErrorConnectionFailed(ResponseMessageError):
    pass


class ErrorConnectionFailedTransientError(ResponseMessageError):
    pass


class ErrorContainsFilterWrongType(ResponseMessageError):
    pass


class ErrorContentConversionFailed(ResponseMessageError):
    pass


class ErrorCorruptData(ResponseMessageError):
    pass


class ErrorCreateItemAccessDenied(ResponseMessageError):
    pass


class ErrorCreateManagedFolderPartialCompletion(ResponseMessageError):
    pass


class ErrorCreateSubfolderAccessDenied(ResponseMessageError):
    pass


class ErrorCrossMailboxMoveCopy(ResponseMessageError):
    pass


class ErrorCrossSiteRequest(ResponseMessageError):
    pass


class ErrorDataSizeLimitExceeded(ResponseMessageError):
    pass


class ErrorDataSourceOperation(ResponseMessageError):
    pass


class ErrorDelegateAlreadyExists(ResponseMessageError):
    pass


class ErrorDelegateCannotAddOwner(ResponseMessageError):
    pass


class ErrorDelegateMissingConfiguration(ResponseMessageError):
    pass


class ErrorDelegateNoUser(ResponseMessageError):
    pass


class ErrorDelegateValidationFailed(ResponseMessageError):
    pass


class ErrorDeleteDistinguishedFolder(ResponseMessageError):
    pass


class ErrorDeleteItemsFailed(ResponseMessageError):
    pass


class ErrorDistinguishedUserNotSupported(ResponseMessageError):
    pass


class ErrorDistributionListMemberNotExist(ResponseMessageError):
    pass


class ErrorDuplicateInputFolderNames(ResponseMessageError):
    pass


class ErrorDuplicateSOAPHeader(ResponseMessageError):
    pass


class ErrorDuplicateUserIdsSpecified(ResponseMessageError):
    pass


class ErrorEmailAddressMismatch(ResponseMessageError):
    pass


class ErrorEventNotFound(ResponseMessageError):
    pass


class ErrorExceededConnectionCount(ResponseMessageError):
    pass


class ErrorExceededFindCountLimit(ResponseMessageError):
    pass


class ErrorExceededSubscriptionCount(ResponseMessageError):
    pass


class ErrorExpiredSubscription(ResponseMessageError):
    pass


class ErrorFolderCorrupt(ResponseMessageError):
    pass


class ErrorFolderExists(ResponseMessageError):
    pass


class ErrorFolderNotFound(ResponseMessageError):
    pass


class ErrorFolderPropertyRequestFailed(ResponseMessageError):
    pass


class ErrorFolderSave(ResponseMessageError):
    pass


class ErrorFolderSaveFailed(ResponseMessageError):
    pass


class ErrorFolderSavePropertyError(ResponseMessageError):
    pass


class ErrorFreeBusyDLLimitReached(ResponseMessageError):
    pass


class ErrorFreeBusyGenerationFailed(ResponseMessageError):
    pass


class ErrorGetServerSecurityDescriptorFailed(ResponseMessageError):
    pass


class ErrorImpersonateUserDenied(ResponseMessageError):
    pass


class ErrorImpersonationDenied(ResponseMessageError):
    pass


class ErrorImpersonationFailed(ResponseMessageError):
    pass


class ErrorInboxRulesValidationError(ResponseMessageError):
    pass


class ErrorIncorrectSchemaVersion(ResponseMessageError):
    pass


class ErrorIncorrectUpdatePropertyCount(ResponseMessageError):
    pass


class ErrorIndividualMailboxLimitReached(ResponseMessageError):
    pass


class ErrorInsufficientResources(ResponseMessageError):
    pass


class ErrorInternalServerError(ResponseMessageError):
    pass


class ErrorInternalServerTransientError(ResponseMessageError):
    pass


class ErrorInvalidAccessLevel(ResponseMessageError):
    pass


class ErrorInvalidArgument(ResponseMessageError):
    pass


class ErrorInvalidAttachmentId(ResponseMessageError):
    pass


class ErrorInvalidAttachmentSubfilter(ResponseMessageError):
    pass


class ErrorInvalidAttachmentSubfilterTextFilter(ResponseMessageError):
    pass


class ErrorInvalidAuthorizationContext(ResponseMessageError):
    pass


class ErrorInvalidChangeKey(ResponseMessageError):
    pass


class ErrorInvalidClientSecurityContext(ResponseMessageError):
    pass


class ErrorInvalidCompleteDate(ResponseMessageError):
    pass


class ErrorInvalidContactEmailAddress(ResponseMessageError):
    pass


class ErrorInvalidContactEmailIndex(ResponseMessageError):
    pass


class ErrorInvalidCrossForestCredentials(ResponseMessageError):
    pass


class ErrorInvalidDelegatePermission(ResponseMessageError):
    pass


class ErrorInvalidDelegateUserId(ResponseMessageError):
    pass


class ErrorInvalidExchangeImpersonationHeaderData(ResponseMessageError):
    pass


class ErrorInvalidExcludesRestriction(ResponseMessageError):
    pass


class ErrorInvalidExpressionTypeForSubFilter(ResponseMessageError):
    pass


class ErrorInvalidExtendedProperty(ResponseMessageError):
    pass


class ErrorInvalidExtendedPropertyValue(ResponseMessageError):
    pass


class ErrorInvalidExternalSharingInitiator(ResponseMessageError):
    pass


class ErrorInvalidExternalSharingSubscriber(ResponseMessageError):
    pass


class ErrorInvalidFederatedOrganizationId(ResponseMessageError):
    pass


class ErrorInvalidFolderId(ResponseMessageError):
    pass


class ErrorInvalidFolderTypeForOperation(ResponseMessageError):
    pass


class ErrorInvalidFractionalPagingParameters(ResponseMessageError):
    pass


class ErrorInvalidFreeBusyViewType(ResponseMessageError):
    pass


class ErrorInvalidGetSharingFolderRequest(ResponseMessageError):
    pass


class ErrorInvalidId(ResponseMessageError):
    pass


class ErrorInvalidIdEmpty(ResponseMessageError):
    pass


class ErrorInvalidIdMalformed(ResponseMessageError):
    pass


class ErrorInvalidIdMalformedEwsLegacyIdFormat(ResponseMessageError):
    pass


class ErrorInvalidIdMonikerTooLong(ResponseMessageError):
    pass


class ErrorInvalidIdNotAnItemAttachmentId(ResponseMessageError):
    pass


class ErrorInvalidIdReturnedByResolveNames(ResponseMessageError):
    pass


class ErrorInvalidIdStoreObjectIdTooLong(ResponseMessageError):
    pass


class ErrorInvalidIdTooManyAttachmentLevels(ResponseMessageError):
    pass


class ErrorInvalidIdXml(ResponseMessageError):
    pass


class ErrorInvalidIndexedPagingParameters(ResponseMessageError):
    pass


class ErrorInvalidInternetHeaderChildNodes(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationAcceptItem(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationCancelItem(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationCreateItem(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationCreateItemAttachment(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationDeclineItem(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationExpandDL(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationRemoveItem(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationSendItem(ResponseMessageError):
    pass


class ErrorInvalidItemForOperationTentative(ResponseMessageError):
    pass


class ErrorInvalidLicense(ResponseMessageError):
    pass


class ErrorInvalidLogonType(ResponseMessageError):
    pass


class ErrorInvalidMailbox(ResponseMessageError):
    pass


class ErrorInvalidManagedFolderProperty(ResponseMessageError):
    pass


class ErrorInvalidManagedFolderQuota(ResponseMessageError):
    pass


class ErrorInvalidManagedFolderSize(ResponseMessageError):
    pass


class ErrorInvalidMergedFreeBusyInterval(ResponseMessageError):
    pass


class ErrorInvalidNameForNameResolution(ResponseMessageError):
    pass


class ErrorInvalidNetworkServiceContext(ResponseMessageError):
    pass


class ErrorInvalidOofParameter(ResponseMessageError):
    pass


class ErrorInvalidOperation(ResponseMessageError):
    pass


class ErrorInvalidOrganizationRelationshipForFreeBusy(ResponseMessageError):
    pass


class ErrorInvalidPagingMaxRows(ResponseMessageError):
    pass


class ErrorInvalidParentFolder(ResponseMessageError):
    pass


class ErrorInvalidPercentCompleteValue(ResponseMessageError):
    pass


class ErrorInvalidPermissionSettings(ResponseMessageError):
    pass


class ErrorInvalidPhoneCallId(ResponseMessageError):
    pass


class ErrorInvalidPhoneNumber(ResponseMessageError):
    pass


class ErrorInvalidPropertyAppend(ResponseMessageError):
    pass


class ErrorInvalidPropertyDelete(ResponseMessageError):
    pass


class ErrorInvalidPropertyForExists(ResponseMessageError):
    pass


class ErrorInvalidPropertyForOperation(ResponseMessageError):
    pass


class ErrorInvalidPropertyRequest(ResponseMessageError):
    pass


class ErrorInvalidPropertySet(ResponseMessageError):
    pass


class ErrorInvalidPropertyUpdateSentMessage(ResponseMessageError):
    pass


class ErrorInvalidProxySecurityContext(ResponseMessageError):
    pass


class ErrorInvalidPullSubscriptionId(ResponseMessageError):
    pass


class ErrorInvalidPushSubscriptionUrl(ResponseMessageError):
    pass


class ErrorInvalidRecipients(ResponseMessageError):
    pass


class ErrorInvalidRecipientSubfilter(ResponseMessageError):
    pass


class ErrorInvalidRecipientSubfilterComparison(ResponseMessageError):
    pass


class ErrorInvalidRecipientSubfilterOrder(ResponseMessageError):
    pass


class ErrorInvalidRecipientSubfilterTextFilter(ResponseMessageError):
    pass


class ErrorInvalidReferenceItem(ResponseMessageError):
    pass


class ErrorInvalidRequest(ResponseMessageError):
    pass


class ErrorInvalidRestriction(ResponseMessageError):
    pass


class ErrorInvalidRoutingType(ResponseMessageError):
    pass


class ErrorInvalidScheduledOofDuration(ResponseMessageError):
    pass


class ErrorInvalidSchemaVersionForMailboxVersion(ResponseMessageError):
    pass


class ErrorInvalidSecurityDescriptor(ResponseMessageError):
    pass


class ErrorInvalidSendItemSaveSettings(ResponseMessageError):
    pass


class ErrorInvalidSerializedAccessToken(ResponseMessageError):
    pass


class ErrorInvalidServerVersion(ResponseMessageError):
    pass


class ErrorInvalidSharingData(ResponseMessageError):
    pass


class ErrorInvalidSharingMessage(ResponseMessageError):
    pass


class ErrorInvalidSid(ResponseMessageError):
    pass


class ErrorInvalidSIPUri(ResponseMessageError):
    pass


class ErrorInvalidSmtpAddress(ResponseMessageError):
    pass


class ErrorInvalidSubfilterType(ResponseMessageError):
    pass


class ErrorInvalidSubfilterTypeNotAttendeeType(ResponseMessageError):
    pass


class ErrorInvalidSubfilterTypeNotRecipientType(ResponseMessageError):
    pass


class ErrorInvalidSubscription(ResponseMessageError):
    pass


class ErrorInvalidSubscriptionRequest(ResponseMessageError):
    pass


class ErrorInvalidSyncStateData(ResponseMessageError):
    pass


class ErrorInvalidTimeInterval(ResponseMessageError):
    pass


class ErrorInvalidUserInfo(ResponseMessageError):
    pass


class ErrorInvalidUserOofSettings(ResponseMessageError):
    pass


class ErrorInvalidUserPrincipalName(ResponseMessageError):
    pass


class ErrorInvalidUserSid(ResponseMessageError):
    pass


class ErrorInvalidUserSidMissingUPN(ResponseMessageError):
    pass


class ErrorInvalidValueForProperty(ResponseMessageError):
    pass


class ErrorInvalidWatermark(ResponseMessageError):
    pass


class ErrorIPGatewayNotFound(ResponseMessageError):
    pass


class ErrorIrresolvableConflict(ResponseMessageError):
    pass


class ErrorItemCorrupt(ResponseMessageError):
    pass


class ErrorItemNotFound(ResponseMessageError):
    pass


class ErrorItemPropertyRequestFailed(ResponseMessageError):
    pass


class ErrorItemSave(ResponseMessageError):
    pass


class ErrorItemSavePropertyError(ResponseMessageError):
    pass


class ErrorLegacyMailboxFreeBusyViewTypeNotMerged(ResponseMessageError):
    pass


class ErrorLocalServerObjectNotFound(ResponseMessageError):
    pass


class ErrorLogonAsNetworkServiceFailed(ResponseMessageError):
    pass


class ErrorMailboxConfiguration(ResponseMessageError):
    pass


class ErrorMailboxDataArrayEmpty(ResponseMessageError):
    pass


class ErrorMailboxDataArrayTooBig(ResponseMessageError):
    pass


class ErrorMailboxFailover(ResponseMessageError):
    pass


class ErrorMailboxLogonFailed(ResponseMessageError):
    pass


class ErrorMailboxMoveInProgress(ResponseMessageError):
    pass


class ErrorMailboxStoreUnavailable(ResponseMessageError):
    pass


class ErrorMailRecipientNotFound(ResponseMessageError):
    pass


class ErrorMailTipsDisabled(ResponseMessageError):
    pass


class ErrorManagedFolderAlreadyExists(ResponseMessageError):
    pass


class ErrorManagedFolderNotFound(ResponseMessageError):
    pass


class ErrorManagedFoldersRootFailure(ResponseMessageError):
    pass


class ErrorMeetingSuggestionGenerationFailed(ResponseMessageError):
    pass


class ErrorMessageDispositionRequired(ResponseMessageError):
    pass


class ErrorMessageSizeExceeded(ResponseMessageError):
    pass


class ErrorMessageTrackingNoSuchDomain(ResponseMessageError):
    pass


class ErrorMessageTrackingPermanentError(ResponseMessageError):
    pass


class ErrorMessageTrackingTransientError(ResponseMessageError):
    pass


class ErrorMimeContentConversionFailed(ResponseMessageError):
    pass


class ErrorMimeContentInvalid(ResponseMessageError):
    pass


class ErrorMimeContentInvalidBase64String(ResponseMessageError):
    pass


class ErrorMissedNotificationEvents(ResponseMessageError):
    pass


class ErrorMissingArgument(ResponseMessageError):
    pass


class ErrorMissingEmailAddress(ResponseMessageError):
    pass


class ErrorMissingEmailAddressForManagedFolder(ResponseMessageError):
    pass


class ErrorMissingInformationEmailAddress(ResponseMessageError):
    pass


class ErrorMissingInformationReferenceItemId(ResponseMessageError):
    pass


class ErrorMissingInformationSharingFolderId(ResponseMessageError):
    pass


class ErrorMissingItemForCreateItemAttachment(ResponseMessageError):
    pass


class ErrorMissingManagedFolderId(ResponseMessageError):
    pass


class ErrorMissingRecipients(ResponseMessageError):
    pass


class ErrorMissingUserIdInformation(ResponseMessageError):
    pass


class ErrorMoreThanOneAccessModeSpecified(ResponseMessageError):
    pass


class ErrorMoveCopyFailed(ResponseMessageError):
    pass


class ErrorMoveDistinguishedFolder(ResponseMessageError):
    pass


class ErrorNameResolutionMultipleResults(ResponseMessageError):
    pass


class ErrorNameResolutionNoMailbox(ResponseMessageError):
    pass


class ErrorNameResolutionNoResults(ResponseMessageError):
    pass


class ErrorNewEventStreamConnectionOpened(ResponseMessageError):
    pass


class ErrorNoApplicableProxyCASServersAvailable(ResponseMessageError):
    pass


class ErrorNoCalendar(ResponseMessageError):
    pass


class ErrorNoDestinationCASDueToKerberosRequirements(ResponseMessageError):
    pass


class ErrorNoDestinationCASDueToSSLRequirements(ResponseMessageError):
    pass


class ErrorNoDestinationCASDueToVersionMismatch(ResponseMessageError):
    pass


class ErrorNoFolderClassOverride(ResponseMessageError):
    pass


class ErrorNoFreeBusyAccess(ResponseMessageError):
    pass


class ErrorNonExistentMailbox(ResponseMessageError):
    pass


class ErrorNonPrimarySmtpAddress(ResponseMessageError):
    pass


class ErrorNoPropertyTagForCustomProperties(ResponseMessageError):
    pass


class ErrorNoPublicFolderReplicaAvailable(ResponseMessageError):
    pass


class ErrorNoPublicFolderServerAvailable(ResponseMessageError):
    pass


class ErrorNoRespondingCASInDestinationSite(ResponseMessageError):
    pass


class ErrorNotAllowedExternalSharingByPolicy(ResponseMessageError):
    pass


class ErrorNotDelegate(ResponseMessageError):
    pass


class ErrorNotEnoughMemory(ResponseMessageError):
    pass


class ErrorNotSupportedSharingMessage(ResponseMessageError):
    pass


class ErrorObjectTypeChanged(ResponseMessageError):
    pass


class ErrorOccurrenceCrossingBoundary(ResponseMessageError):
    pass


class ErrorOccurrenceTimeSpanTooBig(ResponseMessageError):
    pass


class ErrorOperationNotAllowedWithPublicFolderRoot(ResponseMessageError):
    pass


class ErrorOrganizationNotFederated(ResponseMessageError):
    pass


class ErrorOutlookRuleBlobExists(ResponseMessageError):
    pass


class ErrorParentFolderIdRequired(ResponseMessageError):
    pass


class ErrorParentFolderNotFound(ResponseMessageError):
    pass


class ErrorPasswordChangeRequired(ResponseMessageError):
    pass


class ErrorPasswordExpired(ResponseMessageError):
    pass


class ErrorPermissionNotAllowedByPolicy(ResponseMessageError):
    pass


class ErrorPhoneNumberNotDialable(ResponseMessageError):
    pass


class ErrorPropertyUpdate(ResponseMessageError):
    pass


class ErrorPropertyValidationFailure(ResponseMessageError):
    pass


class ErrorProxiedSubscriptionCallFailure(ResponseMessageError):
    pass


class ErrorProxyCallFailed(ResponseMessageError):
    pass


class ErrorProxyGroupSidLimitExceeded(ResponseMessageError):
    pass


class ErrorProxyRequestNotAllowed(ResponseMessageError):
    pass


class ErrorProxyRequestProcessingFailed(ResponseMessageError):
    pass


class ErrorProxyServiceDiscoveryFailed(ResponseMessageError):
    pass


class ErrorProxyTokenExpired(ResponseMessageError):
    pass


class ErrorPublicFolderRequestProcessingFailed(ResponseMessageError):
    pass


class ErrorPublicFolderServerNotFound(ResponseMessageError):
    pass


class ErrorQueryFilterTooLong(ResponseMessageError):
    pass


class ErrorQuotaExceeded(ResponseMessageError):
    pass


class ErrorReadEventsFailed(ResponseMessageError):
    pass


class ErrorReadReceiptNotPending(ResponseMessageError):
    pass


class ErrorRecoverableItemsAccessDenied(ResponseMessageError):
    pass


class ErrorRecurrenceEndDateTooBig(ResponseMessageError):
    pass


class ErrorRecurrenceHasNoOccurrence(ResponseMessageError):
    pass


class ErrorRemoveDelegatesFailed(ResponseMessageError):
    pass


class ErrorRequestAborted(ResponseMessageError):
    pass


class ErrorRequestStreamTooBig(ResponseMessageError):
    pass


class ErrorRequiredPropertyMissing(ResponseMessageError):
    pass


class ErrorResolveNamesInvalidFolderType(ResponseMessageError):
    pass


class ErrorResolveNamesOnlyOneContactsFolderAllowed(ResponseMessageError):
    pass


class ErrorResponseSchemaValidation(ResponseMessageError):
    pass


class ErrorRestrictionTooComplex(ResponseMessageError):
    pass


class ErrorRestrictionTooLong(ResponseMessageError):
    pass


class ErrorResultSetTooBig(ResponseMessageError):
    pass


class ErrorRulesOverQuota(ResponseMessageError):
    pass


class ErrorSavedItemFolderNotFound(ResponseMessageError):
    pass


class ErrorSchemaValidation(ResponseMessageError):
    pass


class ErrorSearchFolderNotInitialized(ResponseMessageError):
    pass


class ErrorSendAsDenied(ResponseMessageError):
    pass


class ErrorSendMeetingCancellationsRequired(ResponseMessageError):
    pass


class ErrorSendMeetingInvitationsOrCancellationsRequired(ResponseMessageError):
    pass


class ErrorSendMeetingInvitationsRequired(ResponseMessageError):
    pass


class ErrorSentMeetingRequestUpdate(ResponseMessageError):
    pass


class ErrorSentTaskRequestUpdate(ResponseMessageError):
    pass


class ErrorServerBusy(ResponseMessageError):
    def __init__(self, *args, **kwargs):
        self.back_off = kwargs.pop("back_off", None)  # Requested back off value in seconds
        super().__init__(*args, **kwargs)


class ErrorServiceDiscoveryFailed(ResponseMessageError):
    pass


class ErrorSharingNoExternalEwsAvailable(ResponseMessageError):
    pass


class ErrorSharingSynchronizationFailed(ResponseMessageError):
    pass


class ErrorStaleObject(ResponseMessageError):
    pass


class ErrorSubmissionQuotaExceeded(ResponseMessageError):
    pass


class ErrorSubscriptionAccessDenied(ResponseMessageError):
    pass


class ErrorSubscriptionDelegateAccessNotSupported(ResponseMessageError):
    pass


class ErrorSubscriptionNotFound(ResponseMessageError):
    pass


class ErrorSubscriptionUnsubsribed(ResponseMessageError):
    pass


class ErrorSyncFolderNotFound(ResponseMessageError):
    pass


class ErrorTimeIntervalTooBig(ResponseMessageError):
    pass


class ErrorTimeoutExpired(ResponseMessageError):
    pass


class ErrorTimeZone(ResponseMessageError):
    pass


class ErrorToFolderNotFound(ResponseMessageError):
    pass


class ErrorTokenSerializationDenied(ResponseMessageError):
    pass


class ErrorTooManyObjectsOpened(ResponseMessageError):
    pass


class ErrorUnableToGetUserOofSettings(ResponseMessageError):
    pass


class ErrorUnifiedMessagingDialPlanNotFound(ResponseMessageError):
    pass


class ErrorUnifiedMessagingRequestFailed(ResponseMessageError):
    pass


class ErrorUnifiedMessagingServerNotFound(ResponseMessageError):
    pass


class ErrorUnsupportedCulture(ResponseMessageError):
    pass


class ErrorUnsupportedMapiPropertyType(ResponseMessageError):
    pass


class ErrorUnsupportedMimeConversion(ResponseMessageError):
    pass


class ErrorUnsupportedPathForQuery(ResponseMessageError):
    pass


class ErrorUnsupportedPathForSortGroup(ResponseMessageError):
    pass


class ErrorUnsupportedPropertyDefinition(ResponseMessageError):
    pass


class ErrorUnsupportedQueryFilter(ResponseMessageError):
    pass


class ErrorUnsupportedRecurrence(ResponseMessageError):
    pass


class ErrorUnsupportedSubFilter(ResponseMessageError):
    pass


class ErrorUnsupportedTypeForConversion(ResponseMessageError):
    pass


class ErrorUpdateDelegatesFailed(ResponseMessageError):
    pass


class ErrorUpdatePropertyMismatch(ResponseMessageError):
    pass


class ErrorUserNotAllowedByPolicy(ResponseMessageError):
    pass


class ErrorUserNotUnifiedMessagingEnabled(ResponseMessageError):
    pass


class ErrorUserWithoutFederatedProxyAddress(ResponseMessageError):
    pass


class ErrorValueOutOfRange(ResponseMessageError):
    pass


class ErrorVirusDetected(ResponseMessageError):
    pass


class ErrorVirusMessageDeleted(ResponseMessageError):
    pass


class ErrorVoiceMailNotImplemented(ResponseMessageError):
    pass


class ErrorWebRequestInInvalidState(ResponseMessageError):
    pass


class ErrorWin32InteropError(ResponseMessageError):
    pass


class ErrorWorkingHoursSaveFailed(ResponseMessageError):
    pass


class ErrorWorkingHoursXmlMalformed(ResponseMessageError):
    pass


class ErrorWrongServerVersion(ResponseMessageError):
    pass


class ErrorWrongServerVersionDelegate(ResponseMessageError):
    pass


# Microsoft recommends caching the autodiscover data around 24 hours and perform autodiscover
# immediately following certain error responses from EWS. See more at
# http://blogs.msdn.com/b/mstehle/archive/2010/11/09/ews-best-practices-use-autodiscover.aspx

ERRORS_REQUIRING_REAUTODISCOVER = (
    ErrorAutoDiscoverFailed,
    ErrorConnectionFailed,
    ErrorIncorrectSchemaVersion,
    ErrorInvalidCrossForestCredentials,
    ErrorInvalidIdReturnedByResolveNames,
    ErrorInvalidNetworkServiceContext,
    ErrorMailboxMoveInProgress,
    ErrorMailboxMoveInProgress,
    ErrorMailboxStoreUnavailable,
    ErrorNameResolutionNoMailbox,
    ErrorNameResolutionNoResults,
    ErrorNotEnoughMemory,
)



from ..errors import InvalidEnumValue, InvalidTypeError
from ..properties import ID_FORMATS, AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId
from ..util import create_element, set_xml_value
from ..version import EXCHANGE_2007_SP1
from .common import EWSService


class ConvertId(EWSService):
    """Take a list of IDs to convert. Returns a list of converted IDs or exception instances, in the same order as the
    input list.

    MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/convertid-operation
    """

    SERVICE_NAME = "ConvertId"
    supported_from = EXCHANGE_2007_SP1
    cls_map = {cls.response_tag(): cls for cls in (AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId)}

    def call(self, items, destination_format):
        if destination_format not in ID_FORMATS:
            raise InvalidEnumValue("destination_format", destination_format, ID_FORMATS)
        return self._elems_to_objs(
            self._chunked_get_elements(self.get_payload, items=items, destination_format=destination_format)
        )

    def _elem_to_obj(self, elem):
        return self.cls_map[elem.tag].from_xml(elem, account=None)

    def get_payload(self, items, destination_format):
        supported_item_classes = AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId
        payload = create_element(f"m:{self.SERVICE_NAME}", attrs=dict(DestinationFormat=destination_format))
        item_ids = create_element("m:SourceIds")
        for item in items:
            if not isinstance(item, supported_item_classes):
                raise InvalidTypeError("item", item, supported_item_classes)
            set_xml_value(item_ids, item, version=self.protocol.version)
        payload.append(item_ids)
        return payload

    @classmethod
    def _get_elements_in_container(cls, container):
        # We may have other elements in here, e.g. 'ResponseCode'. Filter away those.
        return (
            container.findall(AlternateId.response_tag())
            + container.findall(AlternatePublicFolderId.response_tag())
            + container.findall(AlternatePublicFolderItemId.response_tag())
        )



import abc

from ..fields import FieldPath, IndexedField
from ..properties import FolderId
from ..util import MNS, create_element, set_xml_value
from .common import EWSAccountService, parse_folder_elem, to_item_id


class BaseUpdateService(EWSAccountService, metaclass=abc.ABCMeta):
    """Base class for UpdateFolder and UpdateItem"""

    SET_FIELD_ELEMENT_NAME = None
    DELETE_FIELD_ELEMENT_NAME = None
    CHANGE_ELEMENT_NAME = None
    CHANGES_ELEMENT_NAME = None

    @staticmethod
    def _sorted_fields(target_model, fieldnames):
        # Take a list of fieldnames and return the fields in the order they are mentioned in target_model.FIELDS.
        # Checks that all fieldnames are valid.
        fieldnames_copy = list(fieldnames)
        # Loop over FIELDS and not supported_fields(). Upstream should make sure not to update a non-supported field.
        for f in target_model.FIELDS:
            if f.name in fieldnames_copy:
                fieldnames_copy.remove(f.name)
                yield f
        if fieldnames_copy:
            raise ValueError(f"Field name(s) {fieldnames_copy} are not valid {target_model.__name__!r} fields")

    def _get_value(self, target, field):
        return field.clean(getattr(target, field.name), version=self.account.version)  # Make sure the value is OK

    def _set_field_elems(self, target_model, field, value):
        if isinstance(field, IndexedField):
            # Generate either set or delete elements for all combinations of labels and subfields
            supported_labels = field.value_cls.get_field_by_fieldname("label").supported_choices(
                version=self.account.version
            )
            seen_labels = set()
            subfields = field.value_cls.supported_fields(version=self.account.version)
            for v in value:
                seen_labels.add(v.label)
                for subfield in subfields:
                    field_path = FieldPath(field=field, label=v.label, subfield=subfield)
                    subfield_value = getattr(v, subfield.name)
                    if not subfield_value:
                        # Generate delete elements for blank subfield values
                        yield self._delete_field_elem(field_path=field_path)
                    else:
                        # Generate set elements for non-null subfield values
                        yield self._set_field_elem(
                            target_model=target_model,
                            field_path=field_path,
                            value=field.value_cls(**{"label": v.label, subfield.name: subfield_value}),
                        )
                # Generate delete elements for all subfields of all labels not mentioned in the list of values
                for label in (label for label in supported_labels if label not in seen_labels):
                    for subfield in subfields:
                        yield self._delete_field_elem(field_path=FieldPath(field=field, label=label, subfield=subfield))
        else:
            yield self._set_field_elem(target_model=target_model, field_path=FieldPath(field=field), value=value)

    def _set_field_elem(self, target_model, field_path, value):
        set_field = create_element(self.SET_FIELD_ELEMENT_NAME)
        set_xml_value(set_field, field_path, version=self.account.version)
        folder = create_element(target_model.request_tag())
        field_elem = field_path.field.to_xml(value, version=self.account.version)
        set_xml_value(folder, field_elem, version=self.account.version)
        set_field.append(folder)
        return set_field

    def _delete_field_elems(self, field):
        for field_path in FieldPath(field=field).expand(version=self.account.version):
            yield self._delete_field_elem(field_path=field_path)

    def _delete_field_elem(self, field_path):
        delete_folder_field = create_element(self.DELETE_FIELD_ELEMENT_NAME)
        return set_xml_value(delete_folder_field, field_path, version=self.account.version)

    def _update_elems(self, target, fieldnames):
        target_model = target.__class__

        for field in self._sorted_fields(target_model=target_model, fieldnames=fieldnames):
            if field.is_read_only:
                raise ValueError(f"{field.name!r} is a read-only field")
            value = self._get_value(target, field)

            if value is None or (field.is_list and not value):
                # A value of None or [] means we want to remove this field from the item
                if field.is_required or field.is_required_after_save:
                    raise ValueError(f"{field.name!r} is a required field and may not be deleted")
                yield from self._delete_field_elems(field)
                continue

            yield from self._set_field_elems(target_model=target_model, field=field, value=value)

    def _change_elem(self, target, fieldnames):
        if not fieldnames:
            raise ValueError("'fieldnames' must not be empty")
        change = create_element(self.CHANGE_ELEMENT_NAME)
        set_xml_value(change, self._target_elem(target), version=self.account.version)
        updates = create_element("t:Updates")
        for elem in self._update_elems(target=target, fieldnames=fieldnames):
            updates.append(elem)
        change.append(updates)
        return change

    @abc.abstractmethod
    def _target_elem(self, target):
        """Convert the object to update to an XML element"""

    def _changes_elem(self, target_changes):
        changes = create_element(self.CHANGES_ELEMENT_NAME)
        for target, fieldnames in target_changes:
            if not target.account:
                target.account = self.account
            changes.append(self._change_elem(target=target, fieldnames=fieldnames))
        return changes


class UpdateFolder(BaseUpdateService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/updatefolder-operation"""

    SERVICE_NAME = "UpdateFolder"
    SET_FIELD_ELEMENT_NAME = "t:SetFolderField"
    DELETE_FIELD_ELEMENT_NAME = "t:DeleteFolderField"
    CHANGE_ELEMENT_NAME = "t:FolderChange"
    CHANGES_ELEMENT_NAME = "m:FolderChanges"
    element_container_name = f"{{{MNS}}}Folders"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.folders = []  # A hack to communicate parsing args to _elems_to_objs()

    def call(self, folders):
        # We can't easily find the correct folder class from the returned XML. Instead, return objects with the same
        # class as the folder instance it was requested with.
        self.folders = list(folders)  # Convert to a list, in case 'folders' is a generator. We're iterating twice.
        return self._elems_to_objs(self._chunked_get_elements(self.get_payload, items=self.folders))

    def _elems_to_objs(self, elems):
        for (folder, _), elem in zip(self.folders, elems):
            if isinstance(elem, Exception):
                yield elem
                continue
            yield parse_folder_elem(elem=elem, folder=folder, account=self.account)

    def _target_elem(self, target):
        return to_item_id(target, FolderId)

    def get_payload(self, folders):
        # Takes a list of (Folder, fieldnames) tuples where 'Folder' is a instance of a subclass of Folder and
        # 'fieldnames' are the attribute names that were updated.
        payload = create_element(f"m:{self.SERVICE_NAME}")
        payload.append(self._changes_elem(target_changes=folders))
        return payload



from ..properties import TimeZoneDefinition
from ..util import MNS, create_element, peek, set_xml_value
from ..version import EXCHANGE_2010
from .common import EWSService


class GetServerTimeZones(EWSService):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/getservertimezones-operation
    """

    SERVICE_NAME = "GetServerTimeZones"
    element_container_name = f"{{{MNS}}}TimeZoneDefinitions"
    supported_from = EXCHANGE_2010

    def call(self, timezones=None, return_full_timezone_data=False):
        return self._elems_to_objs(
            self._get_elements(
                payload=self.get_payload(timezones=timezones, return_full_timezone_data=return_full_timezone_data)
            )
        )

    def get_payload(self, timezones, return_full_timezone_data):
        payload = create_element(
            f"m:{self.SERVICE_NAME}",
            attrs=dict(ReturnFullTimeZoneData=return_full_timezone_data),
        )
        if timezones is not None:
            is_empty, timezones = peek(timezones)
            if not is_empty:
                tz_ids = create_element("m:Ids")
                for timezone in timezones:
                    tz_id = set_xml_value(create_element("t:Id"), timezone.ms_id)
                    tz_ids.append(tz_id)
                payload.append(tz_ids)
        return payload

    def _elem_to_obj(self, elem):
        return TimeZoneDefinition.from_xml(elem=elem, account=None)



from ..errors import ErrorNameResolutionMultipleResults
from ..properties import Mailbox
from ..util import MNS, create_element, set_xml_value
from .common import EWSService


class ExpandDL(EWSService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/expanddl-operation"""

    SERVICE_NAME = "ExpandDL"
    element_container_name = f"{{{MNS}}}DLExpansion"
    WARNINGS_TO_IGNORE_IN_RESPONSE = ErrorNameResolutionMultipleResults

    def call(self, distribution_list):
        return self._elems_to_objs(self._get_elements(payload=self.get_payload(distribution_list=distribution_list)))

    def _elem_to_obj(self, elem):
        return Mailbox.from_xml(elem, account=None)

    def get_payload(self, distribution_list):
        return set_xml_value(create_element(f"m:{self.SERVICE_NAME}"), distribution_list, version=self.protocol.version)



import logging
import time
from urllib.parse import urlparse

import dns.name
import dns.resolver
from cached_property import threaded_cached_property

from ..configuration import Configuration
from ..errors import AutoDiscoverCircularRedirect, AutoDiscoverFailed, RedirectError, TransportError, UnauthorizedError
from ..protocol import FailFast, Protocol
from ..transport import AUTH_TYPE_MAP, DEFAULT_HEADERS, GSSAPI, NOAUTH, get_auth_method_from_response
from ..util import (
    CONNECTION_ERRORS,
    TLS_ERRORS,
    DummyResponse,
    ParseError,
    _back_off_if_needed,
    get_domain,
    get_redirect_url,
    post_ratelimited,
)
from .cache import autodiscover_cache
from .properties import Autodiscover
from .protocol import AutodiscoverProtocol

log = logging.getLogger(__name__)

DNS_LOOKUP_ERRORS = (
    dns.name.EmptyLabel,
    dns.resolver.NXDOMAIN,
    dns.resolver.NoAnswer,
    dns.resolver.NoNameservers,
)


def discover(email, credentials=None, auth_type=None, retry_policy=None):
    ad_response, protocol = Autodiscovery(email=email, credentials=credentials).discover()
    protocol.config.auth_typ = auth_type
    protocol.config.retry_policy = retry_policy
    return ad_response, protocol


class SrvRecord:
    """A container for autodiscover-related SRV records in DNS."""

    def __init__(self, priority, weight, port, srv):
        self.priority = priority
        self.weight = weight
        self.port = port
        self.srv = srv

    def __eq__(self, other):
        return all(getattr(self, k) == getattr(other, k) for k in self.__dict__)


class Autodiscovery:
    """Autodiscover is a Microsoft protocol for automatically getting the endpoint of the Exchange server and other
    connection-related settings holding the email address using only the email address, and username and password of the
    user.

    For a description of the protocol implemented, see "Autodiscover for Exchange ActiveSync developers":

    https://docs.microsoft.com/en-us/previous-versions/office/developer/exchange-server-interoperability-guidance/hh352638%28v%3dexchg.140%29

    Descriptions of the steps from the article are provided in their respective methods in this class.

    For a description of how to handle autodiscover error messages, see:

    https://docs.microsoft.com/en-us/exchange/client-developer/exchange-web-services/handling-autodiscover-error-messages

    A tip from the article:
    The client can perform steps 1 through 4 in any order or in parallel to expedite the process, but it must wait for
    responses to finish at each step before proceeding. Given that many organizations prefer to use the URL in step 2 to
    set up the Autodiscover service, the client might try this step first.

    Another possibly newer resource which has not yet been attempted is "Outlook 2016 Implementation of Autodiscover":
    https://support.microsoft.com/en-us/help/3211279/outlook-2016-implementation-of-autodiscover

    WARNING: The autodiscover protocol is very complicated. If you have problems autodiscovering using this
    implementation, start by doing an official test at https://testconnectivity.microsoft.com
    """

    # When connecting to servers that may not be serving the correct endpoint, we should use a retry policy that does
    # not leave us hanging for a long time on each step in the protocol.
    INITIAL_RETRY_POLICY = FailFast()
    RETRY_WAIT = 10  # Seconds to wait before retry on connection errors
    MAX_REDIRECTS = 10  # Maximum number of URL redirects before we give up
    DNS_RESOLVER_KWARGS = {}
    DNS_RESOLVER_ATTRS = {
        "timeout": AutodiscoverProtocol.TIMEOUT,
    }

    def __init__(self, email, credentials=None):
        """

        :param email: The email address to autodiscover
        :param credentials: Credentials with authorization to make autodiscover lookups for this Account
            (Default value = None)
        """
        self.email = email
        self.credentials = credentials
        self._urls_visited = []  # Collects HTTP and Autodiscover redirects
        self._redirect_count = 0
        self._emails_visited = []  # Collects Autodiscover email redirects

    def discover(self):
        self._emails_visited.append(self.email.lower())

        # Check the autodiscover cache to see if we already know the autodiscover service endpoint for this email
        # domain. Use a lock to guard against multiple threads competing to cache information.
        log.debug("Waiting for autodiscover_cache lock")
        with autodiscover_cache:
            log.debug("autodiscover_cache lock acquired")
            cache_key = self._cache_key
            domain = get_domain(self.email)
            if cache_key in autodiscover_cache:
                ad_protocol = autodiscover_cache[cache_key]
                log.debug("Cache hit for key %s: %s", cache_key, ad_protocol.service_endpoint)
                try:
                    ad_response = self._quick(protocol=ad_protocol)
                except AutoDiscoverFailed:
                    # Autodiscover no longer works with this domain. Clear cache and try again after releasing the lock
                    log.debug("AD request failure. Removing cache for key %s", cache_key)
                    del autodiscover_cache[cache_key]
                    ad_response = self._step_1(hostname=domain)
            else:
                # This will cache the result
                log.debug("Cache miss for key %s", cache_key)
                ad_response = self._step_1(hostname=domain)

        log.debug("Released autodiscover_cache_lock")
        if ad_response.redirect_address:
            log.debug("Got a redirect address: %s", ad_response.redirect_address)
            if ad_response.redirect_address.lower() in self._emails_visited:
                raise AutoDiscoverCircularRedirect("We were redirected to an email address we have already seen")

            # Start over, but with the new email address
            self.email = ad_response.redirect_address
            return self.discover()

        # We successfully received a response. Clear the cache of seen emails etc.
        self.clear()
        return self._build_response(ad_response=ad_response)

    def clear(self):
        # This resets cached variables
        self._urls_visited = []
        self._redirect_count = 0
        self._emails_visited = []

    @property
    def _cache_key(self):
        # We may be using multiple different credentials and changing our minds on TLS verification. This key
        # combination should be safe for caching.
        domain = get_domain(self.email)
        return domain, self.credentials

    @threaded_cached_property
    def resolver(self):
        resolver = dns.resolver.Resolver(**self.DNS_RESOLVER_KWARGS)
        for k, v in self.DNS_RESOLVER_ATTRS.items():
            setattr(resolver, k, v)
        return resolver

    def _build_response(self, ad_response):
        if not ad_response.autodiscover_smtp_address:
            # Autodiscover does not always return an email address. In that case, the requesting email should be used
            ad_response.user.autodiscover_smtp_address = self.email

        protocol = Protocol(
            config=Configuration(
                service_endpoint=ad_response.protocol.ews_url,
                credentials=self.credentials,
                version=ad_response.version,
                auth_type=ad_response.protocol.auth_type,
            )
        )
        return ad_response, protocol

    def _quick(self, protocol):
        try:
            r = self._get_authenticated_response(protocol=protocol)
        except TransportError as e:
            raise AutoDiscoverFailed(f"Response error: {e}")
        if r.status_code == 200:
            try:
                ad = Autodiscover.from_bytes(bytes_content=r.content)
            except ParseError as e:
                raise AutoDiscoverFailed(f"Invalid response: {e}")
            else:
                return self._step_5(ad=ad)
        raise AutoDiscoverFailed(f"Invalid response code: {r.status_code}")

    def _redirect_url_is_valid(self, url):
        """Three separate responses can be “Redirect responses”:
        * An HTTP status code (301, 302) with a new URL
        * An HTTP status code of 200, but with a payload XML containing a redirect to a different URL
        * An HTTP status code of 200, but with a payload XML containing a different SMTP address as the target address

        We only handle the HTTP 302 redirects here. We validate the URL received in the redirect response to ensure that
        it does not redirect to non-SSL endpoints or SSL endpoints with invalid certificates, and that the redirect is
        not circular. Finally, we should fail after 10 redirects.

        :param url:
        :return:
        """
        if url.lower() in self._urls_visited:
            log.warning("We have already tried this URL: %s", url)
            return False

        if self._redirect_count >= self.MAX_REDIRECTS:
            log.warning("We reached max redirects at URL: %s", url)
            return False

        # We require TLS endpoints
        if not url.startswith("https://"):
            log.debug("Invalid scheme for URL: %s", url)
            return False

        # Quick test that the endpoint responds and that TLS handshake is OK
        try:
            self._get_unauthenticated_response(url, method="head")
        except TransportError as e:
            log.debug("Response error on redirect URL %s: %s", url, e)
            return False

        self._redirect_count += 1
        return True

    def _get_unauthenticated_response(self, url, method="post"):
        """Get auth type by tasting headers from the server. Do POST requests be default. HEAD is too error prone, and
        some servers are set up to redirect to OWA on all requests except POST to the autodiscover endpoint.

        :param url:
        :param method:  (Default value = 'post')
        :return:
        """
        # We are connecting to untrusted servers here, so take necessary precautions.
        hostname = urlparse(url).netloc
        if not self._is_valid_hostname(hostname):
            # 'requests' is really bad at reporting that a hostname cannot be resolved. Let's check this separately.
            # Don't retry on DNS errors. They will most likely be persistent.
            raise TransportError(f"{hostname!r} has no DNS entry")

        kwargs = dict(
            url=url, headers=DEFAULT_HEADERS.copy(), allow_redirects=False, timeout=AutodiscoverProtocol.TIMEOUT
        )
        if method == "post":
            kwargs["data"] = Autodiscover.payload(email=self.email)
        retry = 0
        t_start = time.monotonic()
        while True:
            _back_off_if_needed(self.INITIAL_RETRY_POLICY.back_off_until)
            log.debug("Trying to get response from %s", url)
            with AutodiscoverProtocol.raw_session(url) as s:
                try:
                    r = getattr(s, method)(**kwargs)
                    r.close()  # Release memory
                    break
                except TLS_ERRORS as e:
                    # Don't retry on TLS errors. They will most likely be persistent.
                    raise TransportError(str(e))
                except CONNECTION_ERRORS as e:
                    r = DummyResponse(url=url, request_headers=kwargs["headers"])
                    total_wait = time.monotonic() - t_start
                    if self.INITIAL_RETRY_POLICY.may_retry_on_error(response=r, wait=total_wait):
                        log.debug("Connection error on URL %s (retry %s, error: %s). Cool down", url, retry, e)
                        # Don't respect the 'Retry-After' header. We don't know if this is a useful endpoint, and we
                        # want autodiscover to be reasonably fast.
                        self.INITIAL_RETRY_POLICY.back_off(self.RETRY_WAIT)
                        retry += 1
                        continue
                    log.debug("Connection error on URL %s: %s", url, e)
                    raise TransportError(str(e))
        try:
            auth_type = get_auth_method_from_response(response=r)
        except UnauthorizedError:
            # Failed to guess the auth type
            auth_type = NOAUTH
        if r.status_code in (301, 302) and "location" in r.headers:
            # Make the redirect URL absolute
            try:
                r.headers["location"] = get_redirect_url(r)
            except TransportError:
                del r.headers["location"]
        return auth_type, r

    def _get_authenticated_response(self, protocol):
        """Get a response by using the credentials provided. We guess the auth type along the way.

        :param protocol:
        :return:
        """
        # Redo the request with the correct auth
        data = Autodiscover.payload(email=self.email)
        headers = DEFAULT_HEADERS.copy()
        session = protocol.get_session()
        if GSSAPI in AUTH_TYPE_MAP and isinstance(session.auth, AUTH_TYPE_MAP[GSSAPI]):
            # https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/pox-autodiscover-request-for-exchange
            headers["X-ClientCanHandle"] = "Negotiate"
        try:
            r, session = post_ratelimited(
                protocol=protocol,
                session=session,
                url=protocol.service_endpoint,
                headers=headers,
                data=data,
                allow_redirects=False,
                stream=False,
            )
            protocol.release_session(session)
        except UnauthorizedError as e:
            # It's entirely possible for the endpoint to ask for login. We should continue if login fails because this
            # isn't necessarily the right endpoint to use.
            raise TransportError(str(e))
        except RedirectError as e:
            r = DummyResponse(url=protocol.service_endpoint, headers={"location": e.url}, status_code=302)
        return r

    def _attempt_response(self, url):
        """Return an (is_valid_response, response) tuple.

        :param url:
        :return:
        """
        self._urls_visited.append(url.lower())
        log.debug("Attempting to get a valid response from %s", url)
        try:
            auth_type, r = self._get_unauthenticated_response(url=url)
            ad_protocol = AutodiscoverProtocol(
                config=Configuration(
                    service_endpoint=url,
                    credentials=self.credentials,
                    auth_type=auth_type,
                    retry_policy=self.INITIAL_RETRY_POLICY,
                )
            )
            if auth_type != NOAUTH:
                r = self._get_authenticated_response(protocol=ad_protocol)
        except TransportError as e:
            log.debug("Failed to get a response: %s", e)
            return False, None
        if r.status_code in (301, 302) and "location" in r.headers:
            redirect_url = get_redirect_url(r)
            if self._redirect_url_is_valid(url=redirect_url):
                # The protocol does not specify this explicitly, but by looking at how testconnectivity.microsoft.com
                # works, it seems that we should follow this URL now and try to get a valid response.
                return self._attempt_response(url=redirect_url)
        if r.status_code == 200:
            try:
                ad = Autodiscover.from_bytes(bytes_content=r.content)
            except ParseError as e:
                log.debug("Invalid response: %s", e)
            else:
                # We got a valid response. Unless this is a URL redirect response, we cache the result
                if ad.response is None or not ad.response.redirect_url:
                    cache_key = self._cache_key
                    log.debug("Adding cache entry for key %s: %s", cache_key, ad_protocol.service_endpoint)
                    autodiscover_cache[cache_key] = ad_protocol
                return True, ad
        return False, None

    def _is_valid_hostname(self, hostname):
        log.debug("Checking if %s can be looked up in DNS", hostname)
        try:
            self.resolver.resolve(f"{hostname}.", "A", lifetime=self.DNS_RESOLVER_ATTRS.get("timeout"))
        except DNS_LOOKUP_ERRORS as e:
            log.debug("DNS A lookup failure: %s", e)
            return False
        return True

    def _get_srv_records(self, hostname):
        """Send a DNS query for SRV entries for the hostname.

        An SRV entry that has been formatted for autodiscovery will have the following format:

            canonical name = mail.example.com.
            service = 8 100 443 webmail.example.com.

        The first three numbers in the service line are: priority, weight, port

        :param hostname:
        :return:
        """
        log.debug("Attempting to get SRV records for %s", hostname)
        records = []
        try:
            answers = self.resolver.resolve(f"{hostname}.", "SRV", lifetime=self.DNS_RESOLVER_ATTRS.get("timeout"))
        except DNS_LOOKUP_ERRORS as e:
            log.debug("DNS SRV lookup failure: %s", e)
            return records
        for rdata in answers:
            try:
                vals = rdata.to_text().strip().rstrip(".").split(" ")
                # Raise ValueError if the first three are not ints, and IndexError if there are less than 4 values
                priority, weight, port, srv = int(vals[0]), int(vals[1]), int(vals[2]), vals[3]
                record = SrvRecord(priority=priority, weight=weight, port=port, srv=srv)
                log.debug("Found SRV record %s ", record)
                records.append(record)
            except (ValueError, IndexError):
                log.debug("Incompatible SRV record for %s (%s)", hostname, rdata.to_text())
        return records

    def _step_1(self, hostname):
        """Perform step 1, where the client sends an Autodiscover request to
        https://example.com/autodiscover/autodiscover.xml and then does one of the following:
            * If the Autodiscover attempt succeeds, the client proceeds to step 5.
            * If the Autodiscover attempt fails, the client proceeds to step 2.

        :param hostname:
        :return:
        """
        url = f"https://{hostname}/Autodiscover/Autodiscover.xml"
        log.info("Step 1: Trying autodiscover on %r with email %r", url, self.email)
        is_valid_response, ad = self._attempt_response(url=url)
        if is_valid_response:
            return self._step_5(ad=ad)
        return self._step_2(hostname=hostname)

    def _step_2(self, hostname):
        """Perform step 2, where the client sends an Autodiscover request to
        https://autodiscover.example.com/autodiscover/autodiscover.xml and then does one of the following:
            * If the Autodiscover attempt succeeds, the client proceeds to step 5.
            * If the Autodiscover attempt fails, the client proceeds to step 3.

        :param hostname:
        :return:
        """
        url = f"https://autodiscover.{hostname}/Autodiscover/Autodiscover.xml"
        log.info("Step 2: Trying autodiscover on %r with email %r", url, self.email)
        is_valid_response, ad = self._attempt_response(url=url)
        if is_valid_response:
            return self._step_5(ad=ad)
        return self._step_3(hostname=hostname)

    def _step_3(self, hostname):
        """Perform step 3, where the client sends an unauth'ed GET method request to
        http://autodiscover.example.com/autodiscover/autodiscover.xml (Note that this is a non-HTTPS endpoint). The
        client then does one of the following:
            * If the GET request returns a 302 redirect response, it gets the redirection URL from the 'Location' HTTP
            header and validates it as described in the "Redirect responses" section. The client then does one of the
            following:
                * If the redirection URL is valid, the client tries the URL and then does one of the following:
                    * If the attempt succeeds, the client proceeds to step 5.
                    * If the attempt fails, the client proceeds to step 4.
                * If the redirection URL is not valid, the client proceeds to step 4.
            * If the GET request does not return a 302 redirect response, the client proceeds to step 4.

        :param hostname:
        :return:
        """
        url = f"http://autodiscover.{hostname}/Autodiscover/Autodiscover.xml"
        log.info("Step 3: Trying autodiscover on %r with email %r", url, self.email)
        try:
            _, r = self._get_unauthenticated_response(url=url, method="get")
        except TransportError:
            r = DummyResponse(url=url)
        if r.status_code in (301, 302) and "location" in r.headers:
            redirect_url = get_redirect_url(r)
            if self._redirect_url_is_valid(url=redirect_url):
                is_valid_response, ad = self._attempt_response(url=redirect_url)
                if is_valid_response:
                    return self._step_5(ad=ad)
                log.debug("Got invalid response")
                return self._step_4(hostname=hostname)
            log.debug("Got invalid redirect URL")
            return self._step_4(hostname=hostname)
        log.debug("Got no redirect URL")
        return self._step_4(hostname=hostname)

    def _step_4(self, hostname):
        """Perform step 4, where the client performs a Domain Name System (DNS) query for an SRV record for
        _autodiscover._tcp.example.com. The query might return multiple records. The client selects only records that
        point to an SSL endpoint and that have the highest priority and weight. One of the following actions then
        occurs:
            * If no such records are returned, the client proceeds to step 6.
            * If records are returned, the application randomly chooses a record in the list and validates the endpoint
              that it points to by following the process described in the "Redirect Response" section. The client then
              does one of the following:
                * If the redirection URL is valid, the client tries the URL and then does one of the following:
                    * If the attempt succeeds, the client proceeds to step 5.
                    * If the attempt fails, the client proceeds to step 6.
                * If the redirection URL is not valid, the client proceeds to step 6.

        :param hostname:
        :return:
        """
        dns_hostname = f"_autodiscover._tcp.{hostname}"
        log.info("Step 4: Trying autodiscover on %r with email %r", dns_hostname, self.email)
        srv_records = self._get_srv_records(dns_hostname)
        try:
            srv_host = _select_srv_host(srv_records)
        except ValueError:
            srv_host = None
        if not srv_host:
            return self._step_6()
        redirect_url = f"https://{srv_host}/Autodiscover/Autodiscover.xml"
        if self._redirect_url_is_valid(url=redirect_url):
            is_valid_response, ad = self._attempt_response(url=redirect_url)
            if is_valid_response:
                return self._step_5(ad=ad)
            log.debug("Got invalid response")
            return self._step_6()
        log.debug("Got invalid redirect URL")
        return self._step_6()

    def _step_5(self, ad):
        """Perform step 5. When a valid Autodiscover request succeeds, the following sequence occurs:
            * If the server responds with an HTTP 302 redirect, the client validates the redirection URL according to
              the process defined in the "Redirect responses" and then does one of the following:
                * If the redirection URL is valid, the client tries the URL and then does one of the following:
                    * If the attempt succeeds, the client repeats step 5 from the beginning.
                    * If the attempt fails, the client proceeds to step 6.
                * If the redirection URL is not valid, the client proceeds to step 6.
            * If the server responds with a valid Autodiscover response, the client does one of the following:
                * If the value of the Action element is "Redirect", the client gets the redirection email address from
                  the Redirect element and then returns to step 1, using this new email address.
                * If the value of the Action element is "Settings", the client has successfully received the requested
                  configuration settings for the specified user. The client does not need to proceed to step 6.

        :param ad:
        :return:
        """
        log.info("Step 5: Checking response")
        if ad.response is None:
            # This is not explicit in the protocol, but let's raise errors here
            ad.raise_errors()

        ad_response = ad.response
        if ad_response.redirect_url:
            log.debug("Got a redirect URL: %s", ad_response.redirect_url)
            # We are diverging a bit from the protocol here. We will never get an HTTP 302 since earlier steps already
            # followed the redirects where possible. Instead, we handle redirect responses here.
            if self._redirect_url_is_valid(url=ad_response.redirect_url):
                is_valid_response, ad = self._attempt_response(url=ad_response.redirect_url)
                if is_valid_response:
                    return self._step_5(ad=ad)
                log.debug("Got invalid response")
                return self._step_6()
            log.debug("Invalid redirect URL")
            return self._step_6()
        # This could be an email redirect. Let outer layer handle this
        return ad_response

    def _step_6(self):
        """Perform step 6. If the client cannot contact the Autodiscover service, the client should ask the user for
        the Exchange server name and use it to construct an Exchange EWS URL. The client should try to use this URL for
        future requests.
        """
        raise AutoDiscoverFailed(
            f"All steps in the autodiscover protocol failed for email {self.email}. If you think this is an error, "
            f"consider doing an official test at https://testconnectivity.microsoft.com"
        )


def _select_srv_host(srv_records):
    """Select the record with the highest priority, that also supports TLS.

    :param srv_records:
    :return:
    """
    best_record = None
    for srv_record in srv_records:
        if srv_record.port != 443:
            log.debug("Skipping SRV record %r (no TLS)", srv_record)
            continue
        # Assume port 443 will serve TLS. If not, autodiscover will probably also be broken for others.
        if best_record is None or best_record.priority < srv_record.priority:
            best_record = srv_record
    if not best_record:
        raise ValueError("No suitable records")
    return best_record.srv



import datetime
import logging

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

import tzlocal

from .errors import InvalidTypeError, NaiveDateTimeNotAllowed, UnknownTimeZone
from .winzone import IANA_TO_MS_TIMEZONE_MAP, MS_TIMEZONE_TO_IANA_MAP

log = logging.getLogger(__name__)


class EWSDate(datetime.date):
    """Extends the normal date implementation to satisfy EWS."""

    __slots__ = "_year", "_month", "_day", "_hashcode"

    def ewsformat(self):
        """ISO 8601 format to satisfy xs:date as interpreted by EWS. Example: 2009-01-15."""
        return self.isoformat()

    def __add__(self, other):
        dt = super().__add__(other)
        if isinstance(dt, self.__class__):
            return dt
        return self.from_date(dt)  # We want to return EWSDate objects

    def __iadd__(self, other):
        return self + other

    def __sub__(self, other):
        dt = super().__sub__(other)
        if isinstance(dt, datetime.timedelta):
            return dt
        if isinstance(dt, self.__class__):
            return dt
        return self.from_date(dt)  # We want to return EWSDate objects

    def __isub__(self, other):
        return self - other

    @classmethod
    def fromordinal(cls, n):
        dt = super().fromordinal(n)
        if isinstance(dt, cls):
            return dt
        return cls.from_date(dt)  # We want to return EWSDate objects

    @classmethod
    def from_date(cls, d):
        if type(d) is not datetime.date:
            raise InvalidTypeError("d", d, datetime.date)
        return cls(d.year, d.month, d.day)

    @classmethod
    def from_string(cls, date_string):
        # Sometimes, we'll receive a date string with timezone information. Not very useful.
        if date_string.endswith("Z"):
            date_fmt = "%Y-%m-%dZ"
        elif ":" in date_string:
            if "+" in date_string:
                date_fmt = "%Y-%m-%d+%H:%M"
            else:
                date_fmt = "%Y-%m-%d-%H:%M"
        else:
            date_fmt = "%Y-%m-%d"
        d = datetime.datetime.strptime(date_string, date_fmt).date()
        if isinstance(d, cls):
            return d
        return cls.from_date(d)  # We want to return EWSDate objects


class EWSDateTime(datetime.datetime):
    """Extends the normal datetime implementation to satisfy EWS."""

    __slots__ = "_year", "_month", "_day", "_hour", "_minute", "_second", "_microsecond", "_tzinfo", "_hashcode"

    def __new__(cls, *args, **kwargs):
        # pylint: disable=arguments-differ

        if len(args) == 8:
            tzinfo = args[7]
        else:
            tzinfo = kwargs.get("tzinfo")
        if isinstance(tzinfo, zoneinfo.ZoneInfo):
            # Don't allow pytz or dateutil timezones here. They are not safe to use as direct input for datetime()
            tzinfo = EWSTimeZone.from_timezone(tzinfo)
        if not isinstance(tzinfo, (EWSTimeZone, type(None))):
            raise InvalidTypeError("tzinfo", tzinfo, EWSTimeZone)
        if len(args) == 8:
            args = args[:7] + (tzinfo,)
        else:
            kwargs["tzinfo"] = tzinfo
        return super().__new__(cls, *args, **kwargs)

    def ewsformat(self):
        """ISO 8601 format to satisfy xs:datetime as interpreted by EWS. Examples:
        * 2009-01-15T13:45:56Z
        * 2009-01-15T13:45:56+01:00
        """
        if not self.tzinfo:
            raise ValueError(f"{self!r} must be timezone-aware")
        if self.tzinfo.key == "UTC":
            if self.microsecond:
                return self.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            return self.strftime("%Y-%m-%dT%H:%M:%SZ")
        return self.isoformat()

    @classmethod
    def from_datetime(cls, d):
        if type(d) is not datetime.datetime:
            raise InvalidTypeError("d", d, datetime.datetime)
        if d.tzinfo is None:
            tz = None
        elif isinstance(d.tzinfo, EWSTimeZone):
            tz = d.tzinfo
        else:
            tz = EWSTimeZone.from_timezone(d.tzinfo)
        return cls(d.year, d.month, d.day, d.hour, d.minute, d.second, d.microsecond, tzinfo=tz)

    def astimezone(self, tz=None):
        if tz is None:
            tz = EWSTimeZone.localzone()
        t = super().astimezone(tz=tz).replace(tzinfo=tz)
        if isinstance(t, self.__class__):
            return t
        return self.from_datetime(t)  # We want to return EWSDateTime objects

    def __add__(self, other):
        t = super().__add__(other)
        if isinstance(t, self.__class__):
            return t
        return self.from_datetime(t)  # We want to return EWSDateTime objects

    def __iadd__(self, other):
        return self + other

    def __sub__(self, other):
        t = super().__sub__(other)
        if isinstance(t, datetime.timedelta):
            return t
        if isinstance(t, self.__class__):
            return t
        return self.from_datetime(t)  # We want to return EWSDateTime objects

    def __isub__(self, other):
        return self - other

    @classmethod
    def from_string(cls, date_string):
        # Parses several common datetime formats and returns timezone-aware EWSDateTime objects
        if date_string.endswith("Z"):
            # UTC datetime
            return super().strptime(date_string, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=UTC)
        if len(date_string) == 19:
            # This is probably a naive datetime. Don't allow this, but signal caller with an appropriate error
            local_dt = super().strptime(date_string, "%Y-%m-%dT%H:%M:%S")
            raise NaiveDateTimeNotAllowed(local_dt)
        # This is probably a datetime value with timezone information. This comes in the form '+/-HH:MM'.
        aware_dt = datetime.datetime.fromisoformat(date_string).astimezone(UTC).replace(tzinfo=UTC)
        if isinstance(aware_dt, cls):
            return aware_dt
        return cls.from_datetime(aware_dt)

    @classmethod
    def fromtimestamp(cls, t, tz=None):
        dt = super().fromtimestamp(t, tz=tz)
        if isinstance(dt, cls):
            return dt
        return cls.from_datetime(dt)  # We want to return EWSDateTime objects

    @classmethod
    def utcfromtimestamp(cls, t):
        dt = super().utcfromtimestamp(t)
        if isinstance(dt, cls):
            return dt
        return cls.from_datetime(dt)  # We want to return EWSDateTime objects

    @classmethod
    def now(cls, tz=None):
        t = super().now(tz=tz)
        if isinstance(t, cls):
            return t
        return cls.from_datetime(t)  # We want to return EWSDateTime objects

    @classmethod
    def utcnow(cls):
        t = super().utcnow()
        if isinstance(t, cls):
            return t
        return cls.from_datetime(t)  # We want to return EWSDateTime objects

    def date(self):
        d = super().date()
        if isinstance(d, EWSDate):
            return d
        return EWSDate.from_date(d)  # We want to return EWSDate objects


class EWSTimeZone(zoneinfo.ZoneInfo):
    """Represents a timezone as expected by the EWS TimezoneContext / TimezoneDefinition XML element, and returned by
    services.GetServerTimeZones.
    """

    IANA_TO_MS_MAP = IANA_TO_MS_TIMEZONE_MAP
    MS_TO_IANA_MAP = MS_TIMEZONE_TO_IANA_MAP

    def __new__(cls, *args, **kwargs):
        try:
            instance = super().__new__(cls, *args, **kwargs)
        except zoneinfo.ZoneInfoNotFoundError as e:
            raise UnknownTimeZone(e.args[0])
        try:
            instance.ms_id = cls.IANA_TO_MS_MAP[instance.key][0]
        except KeyError:
            raise UnknownTimeZone(f"No Windows timezone name found for timezone {instance.key!r}")

        # We don't need the Windows long-format timezone name in long format. It's used in timezone XML elements, but
        # EWS happily accepts empty strings. For a full list of timezones supported by the target server, including
        # long-format names, see output of services.GetServerTimeZones(account.protocol).call()
        instance.ms_name = ""
        return instance

    def __eq__(self, other):
        # Microsoft timezones are less granular than IANA, so an EWSTimeZone created from 'Europe/Copenhagen' may return
        # from the server as 'Europe/Copenhagen'. We're catering for Microsoft here, so base equality on the Microsoft
        # timezone ID.
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ms_id == other.ms_id

    @classmethod
    def from_ms_id(cls, ms_id):
        # Create a timezone instance from a Microsoft timezone ID. This is lossy because there is not a 1:1 translation
        # from MS timezone ID to IANA timezone.
        try:
            return cls(cls.MS_TO_IANA_MAP[ms_id])
        except KeyError:
            if "/" in ms_id:
                # EWS sometimes returns an ID that has a region/location format, e.g. 'Europe/Copenhagen'. Try the
                # string unaltered.
                return cls(ms_id)
            raise UnknownTimeZone(f"Windows timezone ID {ms_id!r} is unknown by CLDR")

    @classmethod
    def from_pytz(cls, tz):
        return cls(tz.zone)

    @classmethod
    def from_datetime(cls, tz):
        """Convert from a standard library `datetime.timezone` instance."""
        return cls(tz.tzname(None))

    @classmethod
    def from_dateutil(cls, tz):
        # Objects returned by dateutil.tz.tzlocal() and dateutil.tz.gettz() are not supported. They
        # don't contain enough information to reliably match them with a CLDR timezone.
        if hasattr(tz, "_filename"):
            key = "/".join(tz._filename.split("/")[-2:])
            return cls(key)
        return cls(tz.tzname(datetime.datetime.now()))

    @classmethod
    def from_zoneinfo(cls, tz):
        return cls(tz.key)

    @classmethod
    def from_timezone(cls, tz):
        # Support multiple tzinfo implementations. We could use isinstance(), but then we'd have to have pytz
        # and dateutil as dependencies for this package.
        tz_module = tz.__class__.__module__.split(".")[0]
        try:
            return {
                cls.__module__.split(".")[0]: lambda z: z,
                "backports": cls.from_zoneinfo,
                "datetime": cls.from_datetime,
                "dateutil": cls.from_dateutil,
                "pytz": cls.from_pytz,
                "zoneinfo": cls.from_zoneinfo,
                "pytz_deprecation_shim": lambda z: cls.from_timezone(z.unwrap_shim()),
            }[tz_module](tz)
        except KeyError:
            raise TypeError(f"Unsupported tzinfo type: {tz!r}")

    @classmethod
    def localzone(cls):
        try:
            tz = tzlocal.get_localzone()
        except zoneinfo.ZoneInfoNotFoundError:
            # Older versions of tzlocal will raise a pytz exception. Let's not depend on pytz just for that.
            raise UnknownTimeZone("Failed to guess local timezone")
        # Handle both old and new versions of tzlocal that may return pytz or zoneinfo objects, respectively
        return cls.from_timezone(tz)

    def fromutc(self, dt):
        t = super().fromutc(dt)
        if isinstance(t, EWSDateTime):
            return t
        return EWSDateTime.from_datetime(t)  # We want to return EWSDateTime objects


UTC = EWSTimeZone("UTC")
UTC_NOW = lambda: EWSDateTime.now(tz=UTC)  # noqa: E731



from ..errors import InvalidEnumValue, InvalidTypeError
from ..folders import BaseFolder
from ..properties import ItemId
from ..util import MNS, TNS, peek, xml_text_to_value
from .common import add_xml_child, item_ids_element
from .sync_folder_hierarchy import SyncFolder


class SyncFolderItems(SyncFolder):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/syncfolderitems-operation
    """

    SERVICE_NAME = "SyncFolderItems"
    SYNC_SCOPES = (
        "NormalItems",
        "NormalAndAssociatedItems",
    )
    # Extra change type
    READ_FLAG_CHANGE = "read_flag_change"
    CHANGE_TYPES = SyncFolder.CHANGE_TYPES + (READ_FLAG_CHANGE,)
    shape_tag = "m:ItemShape"
    last_in_range_name = f"{{{MNS}}}IncludesLastItemInRange"
    change_types_map = SyncFolder.change_types_map
    change_types_map[f"{{{TNS}}}ReadFlagChange"] = READ_FLAG_CHANGE

    def call(self, folder, shape, additional_fields, sync_state, ignore, max_changes_returned, sync_scope):
        self.sync_state = sync_state
        if max_changes_returned is None:
            max_changes_returned = self.page_size
        if not isinstance(max_changes_returned, int):
            raise InvalidTypeError("max_changes_returned", max_changes_returned, int)
        if max_changes_returned <= 0:
            raise ValueError(f"'max_changes_returned' {max_changes_returned} must be a positive integer")
        if sync_scope is not None and sync_scope not in self.SYNC_SCOPES:
            raise InvalidEnumValue("sync_scope", sync_scope, self.SYNC_SCOPES)
        return self._elems_to_objs(
            self._get_elements(
                payload=self.get_payload(
                    folder=folder,
                    shape=shape,
                    additional_fields=additional_fields,
                    sync_state=sync_state,
                    ignore=ignore,
                    max_changes_returned=max_changes_returned,
                    sync_scope=sync_scope,
                )
            )
        )

    def _elem_to_obj(self, elem):
        change_type = self.change_types_map[elem.tag]
        if change_type == self.READ_FLAG_CHANGE:
            item = (
                ItemId.from_xml(elem=elem.find(ItemId.response_tag()), account=self.account),
                xml_text_to_value(elem.find(f"{{{TNS}}}IsRead").text, bool),
            )
        elif change_type == self.DELETE:
            item = ItemId.from_xml(elem=elem.find(ItemId.response_tag()), account=self.account)
        else:
            # We can't find() the element because we don't know which tag to look for. The change element can
            # contain multiple item types, each with their own tag.
            item_elem = elem[0]
            item = BaseFolder.item_model_from_tag(item_elem.tag).from_xml(elem=item_elem, account=self.account)
        return change_type, item

    def get_payload(self, folder, shape, additional_fields, sync_state, ignore, max_changes_returned, sync_scope):
        sync_folder_items = self._partial_get_payload(
            folder=folder, shape=shape, additional_fields=additional_fields, sync_state=sync_state
        )
        is_empty, ignore = (True, None) if ignore is None else peek(ignore)
        if not is_empty:
            sync_folder_items.append(item_ids_element(items=ignore, version=self.account.version, tag="m:Ignore"))
        add_xml_child(sync_folder_items, "m:MaxChangesReturned", max_changes_returned)
        if sync_scope:
            add_xml_child(sync_folder_items, "m:SyncScope", sync_scope)
        return sync_folder_items



from ..util import create_element, set_xml_value
from .common import EWSAccountService


class UpdateUserConfiguration(EWSAccountService):
    """MSDN:
    https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/updateuserconfiguration-operation
    """

    SERVICE_NAME = "UpdateUserConfiguration"
    returns_elements = False

    def call(self, user_configuration):
        return self._get_elements(payload=self.get_payload(user_configuration=user_configuration))

    def get_payload(self, user_configuration):
        return set_xml_value(create_element(f"m:{self.SERVICE_NAME}"), user_configuration, version=self.account.version)



from ..errors import InvalidEnumValue
from ..properties import Notification
from ..transport import wrap
from ..util import MNS, create_element
from .common import EWSService, add_xml_child


class SendNotification(EWSService):
    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/sendnotification

    This service is implemented backwards compared to other services. We use it to parse the XML body of push
    notifications we receive on the callback URL defined in a push subscription, and to create responses to these
    push notifications.
    """

    SERVICE_NAME = "SendNotification"
    OK = "OK"
    UNSUBSCRIBE = "Unsubscribe"
    STATUS_CHOICES = (OK, UNSUBSCRIBE)

    def ok_payload(self):
        return wrap(content=self.get_payload(status=self.OK))

    def unsubscribe_payload(self):
        return wrap(content=self.get_payload(status=self.UNSUBSCRIBE))

    def _elem_to_obj(self, elem):
        return Notification.from_xml(elem=elem, account=None)

    @classmethod
    def _response_tag(cls):
        """Return the name of the element containing the service response."""
        return f"{{{MNS}}}{cls.SERVICE_NAME}"

    @classmethod
    def _get_elements_in_container(cls, container):
        return container.findall(Notification.response_tag())

    def get_payload(self, status):
        if status not in self.STATUS_CHOICES:
            raise InvalidEnumValue("status", status, self.STATUS_CHOICES)
        payload = create_element(f"m:{self.SERVICE_NAME}Result")
        add_xml_child(payload, "m:SubscriptionStatus", status)
        return payload



"""
Implements an Exchange user object and access types. Exchange provides two different ways of granting access for a
login to a specific account. Impersonation is used mainly for service accounts that connect via EWS. Delegate is used
for ad-hoc access e.g. granted manually by the user.
See https://docs.microsoft.com/en-us/exchange/client-developer/exchange-web-services/impersonation-and-ews-in-exchange
"""
import abc
import logging
from threading import RLock

from oauthlib.oauth2 import OAuth2Token

from .errors import InvalidTypeError

log = logging.getLogger(__name__)

IMPERSONATION = "impersonation"
DELEGATE = "delegate"
ACCESS_TYPES = (IMPERSONATION, DELEGATE)


class BaseCredentials(metaclass=abc.ABCMeta):
    """Base for credential storage.

    Establishes a method for refreshing credentials (mostly useful with OAuth, which expires tokens relatively
    frequently) and provides a lock for synchronizing access to the object around refreshes.
    """

    def __init__(self):
        self._lock = RLock()

    @property
    def lock(self):
        return self._lock

    @abc.abstractmethod
    def refresh(self, session):
        """Obtain a new set of valid credentials. This is mostly intended to support OAuth token refreshing, which can
        happen in long- running applications or those that cache access tokens and so might start with a token close to
        expiration.

        :param session: requests session asking for refreshed credentials
        :return:
        """

    def _get_hash_values(self):
        return (getattr(self, k) for k in self.__dict__ if k != "_lock")

    def __eq__(self, other):
        return all(getattr(self, k) == getattr(other, k) for k in self.__dict__ if k != "_lock")

    def __hash__(self):
        return hash(tuple(self._get_hash_values()))

    def __getstate__(self):
        # The lock cannot be pickled
        state = self.__dict__.copy()
        del state["_lock"]
        return state

    def __setstate__(self, state):
        # Restore the lock
        self.__dict__.update(state)
        self._lock = RLock()


class Credentials(BaseCredentials):
    r"""Keeps login info the way Exchange likes it.

    Usernames for authentication are of one of these forms:
    * PrimarySMTPAddress
    * WINDOMAIN\username
    * User Principal Name (UPN)
      password: Clear-text password
    """

    EMAIL = "email"
    DOMAIN = "domain"
    UPN = "upn"

    def __init__(self, username, password):
        super().__init__()
        if username.count("@") == 1:
            self.type = self.EMAIL
        elif username.count("\\") == 1:
            self.type = self.DOMAIN
        else:
            self.type = self.UPN
        self.username = username
        self.password = password

    def refresh(self, session):
        pass

    def __repr__(self):
        return self.__class__.__name__ + repr((self.username, "********"))

    def __str__(self):
        return self.username


class OAuth2Credentials(BaseCredentials):
    """Login info for OAuth 2.0 client credentials authentication, as well as a base for other OAuth 2.0 grant types.

    This is primarily useful for in-house applications accessing data from a single Microsoft account. For applications
    that will access multiple tenants' data, the client credentials flow does not give the application enough
    information to restrict end users' access to the appropriate account. Use OAuth2AuthorizationCodeCredentials and
    the associated auth code grant type for multi-tenant applications.
    """

    def __init__(self, client_id, client_secret, tenant_id=None, identity=None, access_token=None):
        """

        :param client_id: ID of an authorized OAuth application, required for automatic token fetching and refreshing
        :param client_secret: Secret associated with the OAuth application
        :param tenant_id: Microsoft tenant ID of the account to access
        :param identity: An Identity object representing the account that these credentials are connected to.
        :param access_token: Previously-obtained access token, as a dict or an oauthlib.oauth2.OAuth2Token
        """
        super().__init__()
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.identity = identity
        self.access_token = access_token

    def refresh(self, session):
        # Creating a new session gets a new access token, so there's no work here to refresh the credentials. This
        # implementation just makes sure we don't raise a NotImplementedError.
        pass

    def on_token_auto_refreshed(self, access_token):
        """Set the access_token. Called after the access token is refreshed (requests-oauthlib can automatically
        refresh tokens if given an OAuth client ID and secret, so this is how our copy of the token stays up-to-date).
        Applications that cache access tokens can override this to store the new token - just remember to call the
        super() method.

        :param access_token: New token obtained by refreshing
        """
        # Ensure we don't update the object in the middle of a new session being created, which could cause a race.
        if not isinstance(access_token, dict):
            raise InvalidTypeError("access_token", access_token, OAuth2Token)
        with self.lock:
            log.debug("%s auth token for %s", "Refreshing" if self.access_token else "Setting", self.client_id)
            self.access_token = access_token

    def _get_hash_values(self):
        # 'access_token' may be refreshed once in a while. This should not affect the hash signature.
        # 'identity' is just informational and should also not affect the hash signature.
        return (getattr(self, k) for k in self.__dict__ if k not in ("_lock", "identity", "access_token"))

    def sig(self):
        # Like hash(self), but pulls in the access token. Protocol.refresh_credentials() uses this to find out
        # if the access_token needs to be refreshed.
        res = []
        for k in self.__dict__:
            if k in ("_lock", "identity"):
                continue
            if k == "access_token":
                res.append(self.access_token["access_token"] if self.access_token else None)
                continue
            res.append(getattr(self, k))
        return hash(tuple(res))

    def __repr__(self):
        return self.__class__.__name__ + repr((self.client_id, "********"))

    def __str__(self):
        return self.client_id


class OAuth2AuthorizationCodeCredentials(OAuth2Credentials):
    """Login info for OAuth 2.0 authentication using the authorization code grant type. This can be used in one of
    several ways:
    * Given an authorization code, client ID, and client secret, fetch a token ourselves and refresh it as needed if
      supplied with a refresh token.
    * Given an existing access token, client ID, and client secret, use the access token until it expires and then
      refresh it as needed.
    * Given only an existing access token, use it until it expires. This can be used to let the calling application
      refresh tokens itself by subclassing and implementing refresh().

    Unlike the base (client credentials) grant, authorization code credentials don't require a Microsoft tenant ID
    because each access token (and the authorization code used to get the access token) is restricted to a single
    tenant.
    """

    def __init__(self, authorization_code=None, access_token=None, client_id=None, client_secret=None, **kwargs):
        """

        :param client_id: ID of an authorized OAuth application, required for automatic token fetching and refreshing
        :param client_secret: Secret associated with the OAuth application
        :param tenant_id: Microsoft tenant ID of the account to access
        :param identity: An Identity object representing the account that these credentials are connected to.
        :param authorization_code: Code obtained when authorizing the application to access an account. In combination
          with client_id and client_secret, will be used to obtain an access token.
        :param access_token: Previously-obtained access token. If a token exists and the application will handle
          refreshing by itself (or opts not to handle it), this parameter alone is sufficient.
        """
        super().__init__(client_id=client_id, client_secret=client_secret, **kwargs)
        self.authorization_code = authorization_code
        if access_token is not None and not isinstance(access_token, dict):
            raise InvalidTypeError("access_token", access_token, OAuth2Token)
        self.access_token = access_token

    def __repr__(self):
        return self.__class__.__name__ + repr(
            (self.client_id, "[client_secret]", "[authorization_code]", "[access_token]")
        )

    def __str__(self):
        client_id = self.client_id
        credential = (
            "[access_token]"
            if self.access_token is not None
            else ("[authorization_code]" if self.authorization_code is not None else None)
        )
        description = " ".join(filter(None, [client_id, credential]))
        return description or "[underspecified credentials]"



import logging
from copy import deepcopy

from ..errors import ErrorFolderNotFound, ErrorItemNotFound, ErrorNoPublicFolderReplicaAvailable, InvalidTypeError
from ..properties import FolderId, InvalidField
from ..queryset import DoesNotExist, MultipleObjectsReturned
from ..restriction import Q

# Traversal enums
SHALLOW = "Shallow"
SOFT_DELETED = "SoftDeleted"
DEEP = "Deep"
FOLDER_TRAVERSAL_CHOICES = (SHALLOW, DEEP, SOFT_DELETED)

MISSING_FOLDER_ERRORS = (ErrorFolderNotFound, ErrorItemNotFound, ErrorNoPublicFolderReplicaAvailable)


log = logging.getLogger(__name__)


class FolderQuerySet:
    """A QuerySet-like class for finding subfolders of a folder collection."""

    def __init__(self, folder_collection):
        from .collections import FolderCollection

        if not isinstance(folder_collection, FolderCollection):
            raise InvalidTypeError("folder_collection", folder_collection, FolderCollection)
        self.folder_collection = folder_collection
        self.q = Q()  # Default to no restrictions
        self.only_fields = None
        self._depth = None

    def _copy_cls(self):
        return self.__class__(folder_collection=self.folder_collection)

    def _copy_self(self):
        """Chaining operations must make a copy of self before making any modifications."""
        new_qs = self._copy_cls()
        new_qs.q = deepcopy(self.q)
        new_qs.only_fields = self.only_fields
        new_qs._depth = self._depth
        return new_qs

    def only(self, *args):
        """Restrict the fields returned. 'name' and 'folder_class' are always returned."""
        from .base import Folder

        # Subfolders will always be of class Folder
        all_fields = self.folder_collection.get_folder_fields(target_cls=Folder, is_complex=None)
        all_fields.update(Folder.attribute_fields())
        only_fields = []
        for arg in args:
            for field_path in all_fields:
                if field_path.field.name == arg:
                    only_fields.append(field_path)
                    break
            else:
                raise InvalidField(f"Unknown field {arg!r} on folders {self.folder_collection.folders}")
        new_qs = self._copy_self()
        new_qs.only_fields = only_fields
        return new_qs

    def depth(self, depth):
        """Specify the search depth. Possible values are: SHALLOW or DEEP.

        :param depth:
        """
        new_qs = self._copy_self()
        new_qs._depth = depth
        return new_qs

    def get(self, *args, **kwargs):
        """Return the query result as exactly one item. Raises DoesNotExist if there are no results, and
        MultipleObjectsReturned if there are multiple results.
        """
        from .collections import FolderCollection

        if not args and set(kwargs) in ({"id"}, {"id", "changekey"}):
            folders = list(
                FolderCollection(account=self.folder_collection.account, folders=[FolderId(**kwargs)]).resolve()
            )
        elif args or kwargs:
            folders = list(self.filter(*args, **kwargs))
        else:
            folders = list(self.all())
        if not folders:
            raise DoesNotExist("Could not find a child folder matching the query")
        if len(folders) != 1:
            raise MultipleObjectsReturned(f"Expected result length 1, but got {folders}")
        f = folders[0]
        if isinstance(f, Exception):
            raise f
        return f

    def all(self):
        """ """
        new_qs = self._copy_self()
        return new_qs

    def filter(self, *args, **kwargs):
        """Add restrictions to the folder search."""
        new_qs = self._copy_self()
        q = Q(*args, **kwargs)
        new_qs.q = new_qs.q & q
        return new_qs

    def __iter__(self):
        return self._query()

    def _query(self):
        from .base import Folder
        from .collections import FolderCollection

        if self.only_fields is None:
            # Subfolders will always be of class Folder
            non_complex_fields = self.folder_collection.get_folder_fields(target_cls=Folder, is_complex=False)
            complex_fields = self.folder_collection.get_folder_fields(target_cls=Folder, is_complex=True)
        else:
            non_complex_fields = {f for f in self.only_fields if not f.field.is_complex}
            complex_fields = {f for f in self.only_fields if f.field.is_complex}

        # First, fetch all non-complex fields using FindFolder. We do this because some folders do not support
        # GetFolder but we still want to get as much information as possible.
        folders = self.folder_collection.find_folders(q=self.q, depth=self._depth, additional_fields=non_complex_fields)
        if not complex_fields:
            yield from folders
            return

        # Fetch all properties for the found folders
        resolveable_folders = []
        for f in folders:
            if isinstance(f, Exception):
                yield f
                continue
            if not f.get_folder_allowed:
                log.debug("GetFolder not allowed on folder %s. Non-complex fields must be fetched with FindFolder", f)
                yield f
            else:
                resolveable_folders.append(f)

        # Get the complex fields using GetFolder, for the folders that support it, and add the extra field values
        complex_folders = FolderCollection(
            account=self.folder_collection.account, folders=resolveable_folders
        ).get_folders(additional_fields=complex_fields)
        for f, complex_f in zip(resolveable_folders, complex_folders):
            if isinstance(f, MISSING_FOLDER_ERRORS):
                # We were unlucky. The folder disappeared between the FindFolder and the GetFolder calls
                continue
            if isinstance(complex_f, Exception):
                yield complex_f
                continue
            # Add the extra field values to the folders we fetched with find_folders()
            if f.__class__ != complex_f.__class__:
                raise ValueError(f"Type mismatch: {f} vs {complex_f}")
            for complex_field in complex_fields:
                field_name = complex_field.field.name
                setattr(f, field_name, getattr(complex_f, field_name))
            yield f


class SingleFolderQuerySet(FolderQuerySet):
    """A helper class with simpler argument types."""

    def __init__(self, account, folder):
        from .collections import FolderCollection

        folder_collection = FolderCollection(account=account, folders=[folder])
        super().__init__(folder_collection=folder_collection)

    def _copy_cls(self):
        return self.__class__(account=self.folder_collection.account, folder=self.folder_collection.folders[0])

    def resolve(self):
        return list(self.folder_collection.resolve())[0]
